# -*- coding: utf-8 -*-

import torch
from mmv_im2im.utils.misc import parse_config

###############################################################################
# pix2pix HD original
# Total loss = ( GAN loss ) + ( lambda_FM ) * ( Feature Matching loss )
###############################################################################


class pix2pix_HD_original:
    def __init__(self, gan_loss, fm_loss, weights, **kwargs):
        super().__init__()
        self.gan_loss = parse_config(gan_loss)
        self.fm_loss = parse_config(fm_loss)
        self.gan_weight = weights["gan_loss"]
        self.fm_weight = weights["fm_loss"]

    def multi_gan(self, real_features, fake_features, n_D):
        loss_gan = 0
        for i in range(n_D):
            real_grid = torch.ones_like(real_features[i][-1], requires_grad=False)
            loss_gan += self.gan_loss(fake_features[i][-1], real_grid)
        return loss_gan

    def feature_matching(self, real_features, fake_features, n_D):
        loss_fm = 0
        for i in range(n_D):
            for j in range(len(fake_features[0])):
                loss_fm += self.fm_loss(
                    fake_features[i][j], real_features[i][j].detach()
                )
        return loss_fm

    def generator_step(self, discriminator, image_A, image_B, fake_B):
        assert image_A.size() == image_B.size()

        n_D = discriminator.n_D
        real_features = discriminator(torch.cat((image_A, image_B), dim=1))
        fake_features = discriminator(torch.cat((image_A, fake_B), dim=1))

        # GAN loss
        loss_gan = self.multi_gan(real_features, fake_features, n_D)

        # Feature Matching Loss
        loss_fm = self.feature_matching(real_features, fake_features, n_D)

        return (1.0 / n_D) * (self.gan_weight * loss_gan + self.fm_weight * loss_fm)

    def discriminator_step(self, discriminator, image_A, image_B, fake_B):
        # fake_B has been detached!!!!
        # note: here, A and B could have different number of channels
        assert image_A.size()[2:] == image_B.size()[2:]
        loss_D = 0
        n_D = discriminator.n_D
        real_features = discriminator(torch.cat((image_A, image_B), dim=1))
        fake_features = discriminator(torch.cat((image_A, fake_B), dim=1))

        # Note: real_features[i][j] refers to the j-th output from the
        # i-th discriminator. When j = -1, it refers to the final prediction,
        # otherwise refers to intermediate features
        for i in range(n_D):
            real_grid = torch.ones_like(real_features[i][-1], requires_grad=False)
            fake_grid = torch.zeros_like(real_features[i][-1], requires_grad=False)

            loss_D += (
                self.gan_loss(real_features[i][-1], real_grid)
                + self.gan_loss(fake_features[i][-1], fake_grid)
            ) * 0.5

        return loss_D * (1.0 / n_D)


###############################################################################
# modified pix2pix HD
# Total loss = weighted sum of "GAN loss", "Feature Matching loss", and
# "reconstruction loss"
###############################################################################


class pix2pix_HD(pix2pix_HD_original):
    def __init__(self, gan_loss, fm_loss, weights, **kwargs):
        super().__init__(gan_loss, fm_loss, weights)
        self.recon_loss = parse_config(kwargs["reconstruction_loss"])
        self.recon_weight = weights["reconstruction_loss"]

    def generator_step(self, discriminator, image_A, image_B, fake_B):
        # note: A and B could have different numbers of channels
        assert image_A.size()[2:] == image_B.size()[2:]

        n_D = discriminator.n_D
        real_features = discriminator(torch.cat((image_A, image_B), dim=1))
        fake_features = discriminator(torch.cat((image_A, fake_B), dim=1))

        # GAN loss
        loss_gan = self.multi_gan(real_features, fake_features, n_D)

        # Feature Matching Loss
        loss_fm = self.feature_matching(real_features, fake_features, n_D)

        # reconstruction loss
        loss_recon = self.recon_loss(image_B, fake_B)

        return (1.0 / n_D) * (
            self.gan_weight * loss_gan + self.fm_weight * loss_fm
        ) + self.recon_weight * loss_recon


###############################################################################
# original pix2pix basic
# Total loss = weighted sum of "GAN loss", "Feature Matching loss", and
# "reconstruction loss"
###############################################################################


class pix2pix_basic:
    def __init__(self, gan_loss, reconstruction_loss, weights):
        super().__init__()
        self.gan_loss = parse_config(gan_loss)
        self.recon_loss = parse_config(reconstruction_loss)
        self.gan_weight = weights["gan_loss"]
        self.recon_weight = weights["reconstruction_loss"]

    def generator_step(self, discriminator, image_A, image_B, fake_B):
        assert image_A.size() == image_B.size()
        fake_features = discriminator(torch.cat((image_A, fake_B), dim=1))

        # GAN loss
        real_grid = torch.ones_like(fake_features, requires_grad=False)
        loss_gan = self.gan_loss(fake_features, real_grid)

        # reconstruction loss
        loss_recon = self.recon_loss(image_B, fake_B)

        return self.gan_weight * loss_gan + self.recon_weight * loss_recon

    def discriminator_step(self, discriminator, image_A, image_B, fake_B):
        # fake_B has been detached!!!!

        assert image_A.size() == image_B.size()
        real_features = discriminator(torch.cat((image_A, image_B), dim=1))
        fake_features = discriminator(torch.cat((image_A, fake_B), dim=1))

        real_grid = torch.ones_like(real_features, requires_grad=False)
        fake_grid = torch.zeros_like(real_features, requires_grad=False)

        loss_D = (
            self.gan_loss(real_features, real_grid)
            + self.gan_loss(fake_features, fake_grid)
        ) * 0.5

        return loss_D



import torch
import torch.nn as nn
import torch.nn.functional as F


def get_valid_num_groups(channels):
    """Returns a valid number of groups for GroupNorm."""
    for g in [8, 4, 2, 1]:
        if channels % g == 0:
            return g
    return 1


class ConvBlock(nn.Module):
    """Standard 2D/3D Convolutional Block."""

    def __init__(self, in_channels, out_channels, Conv, GroupNorm):
        super().__init__()
        gn_groups1 = get_valid_num_groups(out_channels)
        gn_groups2 = get_valid_num_groups(out_channels)
        self.conv1 = Conv(in_channels, out_channels, kernel_size=3, padding=1)
        self.gn1 = GroupNorm(gn_groups1, out_channels)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv2 = Conv(out_channels, out_channels, kernel_size=3, padding=1)
        self.gn2 = GroupNorm(gn_groups2, out_channels)
        self.relu2 = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.relu1(self.gn1(self.conv1(x)))
        x = self.relu2(self.gn2(self.conv2(x)))
        return x


class Down(nn.Module):
    """Downsampling block (MaxPool + ConvBlock)."""

    def __init__(self, in_channels, out_channels, MaxPool, ConvBlock, Conv, GroupNorm):
        super().__init__()
        self.pool = MaxPool(2)
        self.conv_block = ConvBlock(in_channels, out_channels, Conv, GroupNorm)

    def forward(self, x):
        x = self.pool(x)
        x = self.conv_block(x)
        return x


class Up(nn.Module):
    """Upsampling block (ConvTranspose + Concat + ConvBlock).

    Args:
        in_channels_x1_before_upsample (int): Number of channels of the feature map (x1)
                                                before being upsampled by ConvTranspose.
        in_channels_x2_skip_connection (int): Number of channels of the skip connection (x2).
        out_channels (int): Number of output channels for the final ConvBlock in this Up stage.
    """

    def __init__(
        self,
        in_channels_x1_before_upsample,
        in_channels_x2_skip_connection,
        out_channels,
        ConvTranspose,
        ConvBlock,
        Conv,
        GroupNorm,
        interpolation_mode,
    ):
        super().__init__()
        self.interpolation_mode = interpolation_mode
        self.up = ConvTranspose(
            in_channels_x1_before_upsample,
            in_channels_x1_before_upsample // 2,
            kernel_size=2,
            stride=2,
        )

        channels_for_conv_block = (
            in_channels_x1_before_upsample // 2
        ) + in_channels_x2_skip_connection
        self.conv_block = ConvBlock(
            channels_for_conv_block, out_channels, Conv, GroupNorm
        )

    def forward(self, x1, x2):
        x1 = self.up(x1)

        # Get the spatial size of the skip connection tensor
        spatial_size_x2 = x2.size()[2:]

        if x1.size()[2:] != spatial_size_x2:
            x1 = F.interpolate(
                x1,
                size=spatial_size_x2,
                mode=self.interpolation_mode,
                align_corners=False,
            )

        x = torch.cat([x2, x1], dim=1)
        return self.conv_block(x)


class PriorNet(nn.Module):
    """Network to predict prior distribution (mu, logvar)."""

    def __init__(self, in_channels, latent_dim, Conv):
        super().__init__()
        self.conv = Conv(in_channels, 2 * latent_dim, kernel_size=1)

    def forward(self, x):
        mu_logvar = self.conv(x)
        mu = mu_logvar[:, : self.conv.out_channels // 2, ...]
        logvar = mu_logvar[:, self.conv.out_channels // 2 :, ...]
        return mu, logvar


class PosteriorNet(nn.Module):
    """Network to predict posterior distribution (mu, logvar)."""

    def __init__(self, in_channels, latent_dim, Conv):
        super().__init__()
        self.conv = Conv(in_channels, 2 * latent_dim, kernel_size=1)

    def forward(self, x):
        mu_logvar = self.conv(x)
        mu = mu_logvar[:, : self.conv.out_channels // 2, ...]
        logvar = mu_logvar[:, self.conv.out_channels // 2 :, ...]
        return mu, logvar


class ProbabilisticUNet(nn.Module):
    """Probabilistic UNet model.

    This model can operate in 2D or 3D based on the 'model_type' parameter.
    """

    def __init__(
        self,
        in_channels,
        n_classes,
        latent_dim=6,
        task="segment",
        model_type="2D",
        **kwargs,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.n_classes = n_classes
        self.latent_dim = latent_dim
        self.task = task
        self.model_type = model_type

        # Select the appropriate layers based on model_type
        if model_type == "2D":
            self.Conv = nn.Conv2d
            self.MaxPool = nn.MaxPool2d
            self.ConvTranspose = nn.ConvTranspose2d
            self.GroupNorm = nn.GroupNorm
            self.interpolation_mode = "bilinear"
        elif model_type == "3D":
            self.Conv = nn.Conv3d
            self.MaxPool = nn.MaxPool3d
            self.ConvTranspose = nn.ConvTranspose3d
            self.GroupNorm = nn.GroupNorm
            self.interpolation_mode = "trilinear"
        else:
            raise ValueError(f"Unknown model_type: {model_type}")

        # Encoder path (U-Net)
        self.inc = ConvBlock(in_channels, 32, self.Conv, self.GroupNorm)
        self.down1 = Down(32, 64, self.MaxPool, ConvBlock, self.Conv, self.GroupNorm)
        self.down2 = Down(64, 128, self.MaxPool, ConvBlock, self.Conv, self.GroupNorm)
        self.down3 = Down(128, 256, self.MaxPool, ConvBlock, self.Conv, self.GroupNorm)
        self.down4 = Down(256, 512, self.MaxPool, ConvBlock, self.Conv, self.GroupNorm)

        # Prior and Posterior Networks
        self.prior_net = PriorNet(512, latent_dim, self.Conv)
        # PosteriorNet input channels: 512 (features) + n_classes (one-hot y or float y)
        self.posterior_net = PosteriorNet(512 + n_classes, latent_dim, self.Conv)

        # Decoder Path (U-Net upsampling path)
        self.up1 = Up(
            in_channels_x1_before_upsample=512 + latent_dim,
            in_channels_x2_skip_connection=256,
            out_channels=256,
            ConvTranspose=self.ConvTranspose,
            ConvBlock=ConvBlock,
            Conv=self.Conv,
            GroupNorm=self.GroupNorm,
            interpolation_mode=self.interpolation_mode,
        )

        self.up2 = Up(
            in_channels_x1_before_upsample=256,
            in_channels_x2_skip_connection=128,
            out_channels=128,
            ConvTranspose=self.ConvTranspose,
            ConvBlock=ConvBlock,
            Conv=self.Conv,
            GroupNorm=self.GroupNorm,
            interpolation_mode=self.interpolation_mode,
        )

        self.up3 = Up(
            in_channels_x1_before_upsample=128,
            in_channels_x2_skip_connection=64,
            out_channels=64,
            ConvTranspose=self.ConvTranspose,
            ConvBlock=ConvBlock,
            Conv=self.Conv,
            GroupNorm=self.GroupNorm,
            interpolation_mode=self.interpolation_mode,
        )

        self.up4 = Up(
            in_channels_x1_before_upsample=64,
            in_channels_x2_skip_connection=32,
            out_channels=32,
            ConvTranspose=self.ConvTranspose,
            ConvBlock=ConvBlock,
            Conv=self.Conv,
            GroupNorm=self.GroupNorm,
            interpolation_mode=self.interpolation_mode,
        )

        self.outc = self.Conv(32, n_classes, kernel_size=1)

    def forward(self, x, y=None):
        """
        Forward pass of the Probabilistic UNet.

        Args:
            x (torch.Tensor): Input image tensor (B, C, H, W) for 2D or (B, C, D, H, W) for 3D.
            y (torch.Tensor, optional): Ground truth segmentation mask used for training to
                                         calculate posterior. Defaults to None (for inference).

        Returns:
            tuple: A tuple containing:
                - logits (torch.Tensor): Output logits of the UNet.
                - prior_mu (torch.Tensor): Mean of the prior distribution.
                - prior_logvar (torch.Tensor): Log-variance of the prior distribution.
                - post_mu (torch.Tensor or None): Mean of the posterior distribution (None if y is None).
                - post_logvar (torch.Tensor or None): Log-variance of the posterior distribution (None if y is None).
        """
        # Encoder (U-Net)
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        features = self.down4(x4)  # Bottleneck

        # Prior distribution
        prior_mu, prior_logvar = self.prior_net(features)

        # Posterior calculation and latent variable sampling
        post_mu, post_logvar = None, None
        if y is not None:
            if self.task == "segment":
                # Ensure y is one-hot encoded and downsampled
                y_one_hot = F.one_hot(y.long().squeeze(1), num_classes=self.n_classes)
                y_one_hot = y_one_hot.permute(
                    0, -1, *range(1, y_one_hot.dim() - 1)
                ).float()

                y_downsampled = F.interpolate(
                    y_one_hot, size=features.shape[2:], mode="nearest"
                )
            elif self.task == "regression":
                # For regression, y is already a float tensor, just downsample
                y_downsampled = F.interpolate(
                    y,
                    size=features.shape[2:],
                    mode=self.interpolation_mode,
                    align_corners=False,
                )
            else:
                raise ValueError(f"Unknown task type: {self.task}")

            # Concatenate features and downsampled y for posterior network
            post_mu, post_logvar = self.posterior_net(
                torch.cat([features, y_downsampled], dim=1)
            )

            # Sample 'z' from the posterior distribution
            std_post = torch.exp(0.5 * post_logvar)
            eps = torch.randn_like(std_post)
            z = post_mu + eps * std_post
        else:
            # If 'y' is not provided (inference), sample 'z' from the prior.
            std_prior = torch.exp(0.5 * prior_logvar)
            eps = torch.randn_like(std_prior)
            z = prior_mu + eps * std_prior

        # Expand 'z' to spatial dimensions for concatenation
        spatial_dims = features.size()[2:]
        if z.dim() == 2:  # [B, latent_dim]
            # Expands latent vector to spatial dimensions
            z_expanded = z.unsqueeze(-1)
            for _ in range(len(spatial_dims) - 1):
                z_expanded = z_expanded.unsqueeze(-1)
            z_expanded = z_expanded.repeat(1, 1, *spatial_dims)
        elif z.dim() == 2 + len(spatial_dims):
            if z.size()[2:] != spatial_dims:
                z_expanded = F.interpolate(z, size=spatial_dims, mode="nearest")
            else:
                z_expanded = z
        else:
            raise ValueError(f"Unexpected latent vector z dimension: {z.dim()}")

        # Concatenate bottleneck features with latent vector
        concat_bottleneck = torch.cat([features, z_expanded], dim=1)

        # Decoder (U-Net upsampling path)
        x_up = self.up1(concat_bottleneck, x4)
        x_up = self.up2(x_up, x3)
        x_up = self.up3(x_up, x2)
        x_up = self.up4(x_up, x1)
        output = self.outc(x_up)

        # Return all necessary components for ELBO calculation
        return output, prior_mu, prior_logvar, post_mu, post_logvar



import torch
import torch.nn as nn
import torch.nn.functional as F


class DownsamplerBlock(nn.Module):
    def __init__(self, ninput, noutput):
        super().__init__()

        self.conv = nn.Conv3d(
            ninput, noutput - ninput, (3, 3, 3), stride=2, padding=1, bias=True
        )

        self.pool = nn.MaxPool3d(2, stride=2)

        self.bn = nn.BatchNorm3d(noutput, eps=1e-3)

    def forward(self, input):
        output = torch.cat([self.conv(input), self.pool(input)], 1)
        output = self.bn(output)
        return F.relu(output)


class non_bottleneck_1d(nn.Module):
    def __init__(self, chann, dropprob, dilated):
        super().__init__()
        self.conv3x1x1_1 = nn.Conv3d(
            chann, chann, (3, 1, 1), stride=1, padding=(1, 0, 0), bias=True
        )

        self.conv1x3x1_1 = nn.Conv3d(
            chann, chann, (1, 3, 1), stride=1, padding=(0, 1, 0), bias=True
        )

        self.conv1x1x3_1 = nn.Conv3d(
            chann, chann, (1, 1, 3), stride=1, padding=(0, 0, 1), bias=True
        )

        self.bn1 = nn.BatchNorm3d(chann, eps=1e-03)

        self.conv3x1x1_2 = nn.Conv3d(
            chann,
            chann,
            (3, 1, 1),
            stride=1,
            padding=(1 * dilated, 0, 0),
            bias=True,
            dilation=(dilated, 1, 1),
        )

        self.conv1x3x1_2 = nn.Conv3d(
            chann,
            chann,
            (1, 3, 1),
            stride=1,
            padding=(0, 1 * dilated, 0),
            bias=True,
            dilation=(1, dilated, 1),
        )

        self.conv1x1x3_2 = nn.Conv3d(
            chann,
            chann,
            (1, 1, 3),
            stride=1,
            padding=(0, 0, 1 * dilated),
            bias=True,
            dilation=(1, 1, dilated),
        )

        self.bn2 = nn.BatchNorm3d(chann, eps=1e-03)

        self.dropout = nn.Dropout3d(dropprob)

    def forward(self, input):
        output = self.conv3x1x1_1(input)
        output = F.relu(output)
        output = self.conv1x3x1_1(output)
        output = F.relu(output)
        output = self.conv1x1x3_1(output)
        # TODO think about adding a relu here!

        output = self.bn1(output)
        output = F.relu(output)

        output = self.conv3x1x1_2(output)
        output = F.relu(output)
        output = self.conv1x3x1_2(output)
        output = F.relu(output)
        output = self.conv1x1x3_2(output)
        output = self.bn2(output)

        if self.dropout.p != 0:
            output = self.dropout(output)

        return F.relu(output + input)  # +input = identity (residual connection)


class Encoder(nn.Module):
    def __init__(self, num_classes, input_channels):
        super().__init__()
        self.initial_block = DownsamplerBlock(input_channels, 16)

        self.layers = nn.ModuleList()

        self.layers.append(DownsamplerBlock(16, 64))

        for x in range(0, 5):  # 5 times
            self.layers.append(non_bottleneck_1d(64, 0.03, 1))

        self.layers.append(DownsamplerBlock(64, 128))

        for x in range(0, 2):  # 2 times
            self.layers.append(non_bottleneck_1d(128, 0.3, 2))
            self.layers.append(non_bottleneck_1d(128, 0.3, 4))
            self.layers.append(non_bottleneck_1d(128, 0.3, 8))
            self.layers.append(non_bottleneck_1d(128, 0.3, 16))

        self.output_conv = nn.Conv3d(
            128, num_classes, 1, stride=1, padding=0, bias=True
        )

    def forward(self, input, predict=False):
        output = self.initial_block(input)

        for layer in self.layers:
            output = layer(output)

        if predict:
            output = self.output_conv(output)

        return output


class UpsamplerBlock(nn.Module):
    def __init__(self, ninput, noutput):
        super().__init__()
        self.conv = nn.ConvTranspose3d(
            ninput, noutput, 3, stride=2, padding=1, output_padding=1, bias=True
        )
        self.bn = nn.BatchNorm3d(noutput, eps=1e-3)

    def forward(self, input):
        output = self.conv(input)
        output = self.bn(output)
        return F.relu(output)


class Decoder(nn.Module):
    def __init__(self, num_classes):
        super().__init__()

        self.layers = nn.ModuleList()

        self.layers.append(UpsamplerBlock(128, 64))
        self.layers.append(non_bottleneck_1d(64, 0, 1))
        self.layers.append(non_bottleneck_1d(64, 0, 1))

        self.layers.append(UpsamplerBlock(64, 16))
        self.layers.append(non_bottleneck_1d(16, 0, 1))
        self.layers.append(non_bottleneck_1d(16, 0, 1))

        self.output_conv = nn.ConvTranspose3d(
            16, num_classes, 2, stride=2, padding=0, output_padding=0, bias=True
        )

    def forward(self, input):
        output = input

        for layer in self.layers:
            output = layer(output)

        output = self.output_conv(output)

        return output


# ERFNet
class Net(nn.Module):
    def __init__(
        self, num_classes, input_channels, encoder=None
    ):  # use encoder to pass pretrained encoder
        super().__init__()

        if encoder is None:
            self.encoder = Encoder(num_classes, input_channels)
        else:
            self.encoder = encoder
        self.decoder = Decoder(num_classes)

    def forward(self, input, only_encode=False):
        if only_encode:
            return self.encoder.forward(input, predict=True)
        else:
            output = self.encoder(input)  # predict=False by default
            return self.decoder.forward(output)



from typing import List

import numpy as np
from scipy.ndimage import distance_transform_edt
from skimage.measure import label, regionprops
from skimage.morphology import (
    ball,
    disk,
    dilation,
    erosion,
    medial_axis,
    remove_small_objects,
)


def hole_filling(
    bw: np.ndarray, hole_min: int, hole_max: int, fill_2d: bool = True
) -> np.ndarray:
    """Fill holes in 2D/3D segmentation

    Parameters:
    -------------
    bw: np.ndarray
        a binary 2D/3D image.
    hole_min: int
        the minimum size of the holes to be filled
    hole_max: int
        the maximum size of the holes to be filled
    fill_2d: bool
        if fill_2d=True, a 3D image will be filled slice by slice.
        If you think of a hollow tube alone z direction, the inside
        is not a hole under 3D topology, but the inside on each slice
        is indeed a hole under 2D topology.

    Return:
        a binary image after hole filling
    """
    bw = bw > 0
    if len(bw.shape) == 2:
        background_lab = label(~bw, connectivity=1)
        fill_out = np.copy(background_lab)
        component_sizes = np.bincount(background_lab.ravel())
        too_big = component_sizes > hole_max
        too_big_mask = too_big[background_lab]
        fill_out[too_big_mask] = 0
        too_small = component_sizes < hole_min
        too_small_mask = too_small[background_lab]
        fill_out[too_small_mask] = 0
    elif len(bw.shape) == 3:
        if fill_2d:
            fill_out = np.zeros_like(bw)
            for zz in range(bw.shape[0]):
                background_lab = label(~bw[zz, :, :], connectivity=1)
                out = np.copy(background_lab)
                component_sizes = np.bincount(background_lab.ravel())
                too_big = component_sizes > hole_max
                too_big_mask = too_big[background_lab]
                out[too_big_mask] = 0
                too_small = component_sizes < hole_min
                too_small_mask = too_small[background_lab]
                out[too_small_mask] = 0
                fill_out[zz, :, :] = out
        else:
            background_lab = label(~bw, connectivity=1)
            fill_out = np.copy(background_lab)
            component_sizes = np.bincount(background_lab.ravel())
            too_big = component_sizes > hole_max
            too_big_mask = too_big[background_lab]
            fill_out[too_big_mask] = 0
            too_small = component_sizes < hole_min
            too_small_mask = too_small[background_lab]
            fill_out[too_small_mask] = 0
    else:
        print("error in image shape")
        return

    return np.logical_or(bw, fill_out)


def size_filter(
    img: np.ndarray, min_size: int, method: str = "3D", connectivity: int = 1
):
    """size filter

    Parameters:
    ------------
    img: np.ndarray
        the image to filter on
    min_size: int
        the minimum size to keep
    method: str
        either "3D" or "slice_by_slice", default is "3D"
    connnectivity: int
        the connectivity to use when computing object size
    """
    assert len(img.shape) == 3, "image has to be 3D"
    if method == "3D":
        return remove_small_objects(
            img > 0, min_size=min_size, connectivity=connectivity
        )
    elif method == "slice_by_slice":
        seg = np.zeros(img.shape, dtype=bool)
        for zz in range(img.shape[0]):
            seg[zz, :, :] = remove_small_objects(
                img[zz, :, :] > 0,
                min_size=min_size,
                connectivity=connectivity,
            )
        return seg
    else:
        raise NotImplementedError(f"unsupported method {method}")


def topology_preserving_thinning(
    bw: np.ndarray, min_thickness: int = 1, thin: int = 1
) -> np.ndarray:
    """perform thinning on segmentation without breaking topology

    Parameters:
    --------------
    bw: np.ndarray
        the 3D binary image to be thinned
    min_thickness: int
        Half of the minimum width you want to keep from being thinned.
        For example, when the object width is smaller than 4, you don't
        want to make this part even thinner (may break the thin object
        and alter the topology), you can set this value as 2.
    thin: int
        the amount to thin (has to be an positive integer). The number of
         pixels to be removed from outter boundary towards center.

    Return:
    -------------
        A binary image after thinning
    """
    bw = bw > 0
    safe_zone = np.zeros_like(bw)
    for zz in range(bw.shape[0]):
        if np.any(bw[zz, :, :]):
            ctl = medial_axis(bw[zz, :, :] > 0)
            dist = distance_transform_edt(ctl == 0)
            safe_zone[zz, :, :] = dist > min_thickness + 1e-5

    rm_candidate = np.logical_xor(bw > 0, erosion(bw > 0, ball(thin)))

    bw[np.logical_and(safe_zone, rm_candidate)] = 0

    return bw


def divide_nonzero(array1, array2):
    """
    Divides two arrays. Returns zero when dividing by zero.
    """
    denominator = np.copy(array2)
    denominator[denominator == 0] = 1e-10
    return np.divide(array1, denominator)


def histogram_otsu(hist):
    """Apply Otsu thresholding method on 1D histogram"""

    # modify the elements in hist to avoid completely zero value in cumsum
    hist = hist + 1e-5

    bin_size = 1 / (len(hist) - 1)
    bin_centers = np.arange(0, 1 + 0.5 * bin_size, bin_size)
    hist = hist.astype(float)

    # class probabilities for all possible thresholds
    weight1 = np.cumsum(hist)
    weight2 = np.cumsum(hist[::-1])[::-1]
    # class means for all possible thresholds

    mean1 = np.cumsum(hist * bin_centers) / weight1
    mean2 = (np.cumsum((hist * bin_centers)[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of `weight1`/`mean1` should pair with zero values in
    # `weight2`/`mean2`, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    idx = np.argmax(variance12)
    threshold = bin_centers[:-1][idx]
    return threshold


def absolute_eigenvaluesh(nd_array):
    """Computes the eigenvalues sorted by absolute value from the symmetrical matrix.

    Parameters:
    -------------
    nd_array: nd.ndarray
        array from which the eigenvalues will be calculated.

    Return:
    -------------
        A list with the eigenvalues sorted in absolute ascending order (e.g.
        [eigenvalue1, eigenvalue2, ...])
    """
    eigenvalues = np.linalg.eigvalsh(nd_array)
    sorted_eigenvalues = sortbyabs(eigenvalues, axis=-1)
    return [
        np.squeeze(eigenvalue, axis=-1)
        for eigenvalue in np.split(
            sorted_eigenvalues, sorted_eigenvalues.shape[-1], axis=-1
        )
    ]


def sortbyabs(a: np.ndarray, axis=0):
    """Sort array along a given axis by the absolute value
    modified from: http://stackoverflow.com/a/11253931/4067734
    """
    index = list(np.ix_(*[np.arange(i) for i in a.shape]))
    index[axis] = np.abs(a).argsort(axis)
    return a[tuple(index)]


def get_middle_frame(struct_img: np.ndarray, method: str = "z") -> int:
    """find the middle z frame of an image stack

    Parameters:
    ------------
    struct_img: np.ndarray
        the 3D image to process
    method: str
        which method to use to determine the middle frame. Options
        are "z" or "intensity". "z" is solely based on the number of z
        frames. "intensity" method uses Otsu threshod to estimate the
        volume of foreground signals in the stack, then estimated volume
        of each z frame forms a z-profile, and finally another Otsu
        method is apply on the z profile to find the best z frame (with
        an assumption of two peaks along z profile, one near the bottom
        of the cells and one near the bottom of the cells, so the optimal
        separation is the middle of the stack).

    Return:
    -----------
    mid_frame: int
        the z index of the middle z frame
    """

    from skimage.filters import threshold_otsu

    if method == "intensity":
        bw = struct_img > threshold_otsu(struct_img)
        z_profile = np.zeros((bw.shape[0],), dtype=int)
        for zz in range(bw.shape[0]):
            z_profile[zz] = np.count_nonzero(bw[zz, :, :])
        mid_frame = None
        if isinstance(round(histogram_otsu(z_profile) * bw.shape[0]), int):
            mid_frame = round(histogram_otsu(z_profile) * bw.shape[0])
        else:
            mid_frame = round(histogram_otsu(z_profile) * bw.shape[0]).astype(int)

    elif method == "z":
        mid_frame = struct_img.shape[0] // 2

    else:
        print("unsupported method")
        quit()

    return mid_frame


def get_3dseed_from_mid_frame(
    bw: np.ndarray,
    stack_shape: List = None,
    mid_frame: int = -1,
    hole_min: int = 1,
    bg_seed: bool = True,
) -> np.ndarray:
    """build a 3D seed image from the binary segmentation of a single slice

    Parameters:
    ------------
    bw: np.ndarray
        the 2d segmentation of a single frame, or a 3D array with only one slice
        containing segmentation
    stack_shape: List
        (only used when bw is 2d) the shape of original 3d image, e.g.
        shape_3d = img.shape
    frame_index: int
        (only used when bw is 2d) the index of where bw is from the whole z-stack
    hole_min: int
        any connected component in bw2d with size smaller than area_min
        will be excluded from seed image generation
    bg_seed: bool
        bg_seed=True will add a background seed at the first frame (z=0).

    """
    from skimage.morphology import remove_small_objects

    out = remove_small_objects(bw > 0, hole_min)

    out1 = label(out)
    stat = regionprops(out1)

    # build the seed
    seed = np.zeros(stack_shape)
    seed_count = 0
    if bg_seed:
        seed[0, :, :] = 1
        seed_count += 1

    for idx in range(len(stat)):
        py, px = np.round(stat[idx].centroid)
        seed_count += 1
        seed[mid_frame, int(py), int(px)] = seed_count

    return seed


def remove_hot_pixel(seg: np.ndarray) -> np.ndarray:
    """
    remove hot pixel from segmentation
    """

    assert len(seg.shape) == 3, "input segmentation must be 3D"

    # make sure the segmentation is 0/1
    seg = seg.astype(np.uint8)
    seg[seg > 0] = 1

    # get sum projection along z
    seg_proj = np.sum(seg, axis=0)

    # find hot pixels
    hot_pixel = seg_proj >= seg.shape[0] - 2

    # dilate the area to cover the surrounding pixels
    hot_pixel = dilation(hot_pixel, disk(2))

    # clean up every z
    for z in range(seg.shape[0]):
        seg_z = seg[z, :, :]
        seg_z[hot_pixel] = 0
        seg[z, :, :] = seg_z

    return seg


def get_seed_for_objects(
    raw: np.ndarray,
    bw: np.ndarray,
    area_min: int = 1,
    area_max: int = 10000,
    bg_seed: bool = True,
) -> np.ndarray:
    """
    build a seed image for an image of 3D objects (assuming roughly convex shape
    in 3D) using the information in the middle slice

    Parameters:
    ------------
    raw: np.ndarray
        orignal image used to determine middle slice
    bw: np.ndarray
        a round 3D segmentation, expecting the segmentation in the middle slice
        having relatively good quality
    area_min: int
        estimated minimal size on one single slice (major body chunk, e.g. the
        center XY plane of a 3D ball) of an object
    area_max: int
        estimated maximal size on one single slice (major body chunk, e.g. the
        center XY plane of a 3D ball) of an object. It is recommended to be
        conservertive (setting this value a little larger)
    bg_seed: bool
        bg_seed=True will add a background seed at the first frame (z=0).

    """
    from skimage.morphology import remove_small_objects

    # determine middle slice
    mid_z = get_middle_frame(raw, method="intensity")

    # take seg of middle slice
    bw2d = bw[mid_z, :, :]

    # fillin holes to form solid objects
    bw2d_fill = hole_filling(bw2d, area_min, area_max)

    # prune the objects in middle slice
    out = remove_small_objects(bw2d_fill > 0, area_min)

    # extract object and calculate centroid
    out1 = label(out)
    stat = regionprops(out1)

    # use each centroid as one seed
    seed = np.zeros(raw.shape)
    seed_count = 0
    if bg_seed:
        seed[0, :, :] = 1
        seed_count += 1

    for idx in range(len(stat)):
        py, px = np.round(stat[idx].centroid)
        seed_count += 1
        seed[mid_z, int(py), int(px)] = seed_count

    return seed.astype(int)


def segmentation_union(seg: List) -> np.ndarray:
    """merge multiple segmentations into a single result

    Parameters
    ------------
    seg: List
        a list of segmentations, should all have the same shape
    """

    return np.logical_or.reduce(seg)


def segmentation_intersection(seg: List) -> np.ndarray:
    """get the intersection of multiple segmentations into a single result

    Parameters
    ------------
    seg: List
        a list of segmentations, should all have the same shape
    """

    return np.logical_and.reduce(seg)


def segmentation_xor(seg: List) -> np.ndarray:
    """get the XOR of multiple segmentations into a single result

    Parameters
    ------------
    seg: List
        a list of segmentations, should all have the same shape
    """

    return np.logical_xor.reduce(seg)


def remove_index_object(
    label: np.ndarray, id_to_remove: List[int] = [1], in_place: bool = False
) -> np.ndarray:
    if in_place:
        img = label
    else:
        img = label.copy()

    for id in id_to_remove:
        img[img == id] = 0

    return img


def peak_local_max_wrapper(
    struct_img_for_peak: np.ndarray, bw: np.ndarray
) -> np.ndarray:
    from skimage.feature import peak_local_max

    local_maxi = peak_local_max(struct_img_for_peak, labels=label(bw), min_distance=2)
    local_maxi_image = np.zeros_like(struct_img_for_peak)
    local_maxi_image[tuple(local_maxi.T)] = True
    return local_maxi_image


def watershed_wrapper(bw: np.ndarray, local_maxi: np.ndarray) -> np.ndarray:
    from scipy.ndimage import distance_transform_edt
    from skimage.measure import label
    from skimage.morphology import dilation, ball
    from skimage.segmentation import watershed

    distance = distance_transform_edt(bw)
    im_watershed = watershed(
        -distance,
        label(dilation(local_maxi, footprint=ball(1))),
        mask=bw,
        watershed_line=True,
    )
    return im_watershed


def prune_z_slices(bw: np.ndarray):
    """
    prune the segmentation by only keep a certain range of z-slices
    with the assumption of all signals living only in a few consecutive
    z-slices. This function will first determine the key z-slice where most
    of the signals living on and then include a few slices up/down along z
    to make the segmentation completed. This is useful when you have prior
    knowledge about your segmentation target and can effectively exclude
    small segmented objects due to noise/artifacts in those z-slices we are
    sure the signal should not live on.

    Parameters:
    -----------
    bw: np.ndarray
        the segmentation before pruning
    """
    bw_z = np.zeros(bw.shape[0], dtype=np.uint16)
    for zz in range(bw.shape[0]):
        bw_z[zz] = np.count_nonzero(bw[zz, :, :] > 0)

    mid_z = np.argmax(bw_z)
    low_z = 0
    high_z = bw.shape[0] - 2
    for ii in np.arange(mid_z - 1, 0, -1):
        if bw_z[ii] < 100:
            low_z = ii
            break
    for ii in range(mid_z + 1, bw.shape[0] - 1, 1):
        if bw_z[ii] < 100:
            high_z = ii
            break

    seg = bw.copy()
    seg[:low_z, :, :] = 0
    seg[high_z + 1 :, :, :] = 0

    return seg


def cell_local_adaptive_threshold(
    structure_img_smooth: np.ndarray, cell_wise_min_area: int
):
    from skimage.filters import threshold_triangle, threshold_otsu
    from skimage.morphology import dilation

    # cell-wise local adaptive thresholding
    th_low_level = threshold_triangle(structure_img_smooth)

    bw_low_level = structure_img_smooth > th_low_level
    bw_low_level = remove_small_objects(
        bw_low_level, min_size=cell_wise_min_area, connectivity=1, out=bw_low_level
    )
    bw_low_level = dilation(bw_low_level, footprint=ball(2))

    bw_high_level = np.zeros_like(bw_low_level)
    lab_low, num_obj = label(bw_low_level, return_num=True, connectivity=1)

    for idx in range(num_obj):
        single_obj = lab_low == (idx + 1)
        local_otsu = threshold_otsu(structure_img_smooth[single_obj > 0])
        bw_high_level[
            np.logical_and(structure_img_smooth > local_otsu * 0.98, single_obj)
        ] = 1
    return bw_high_level


def invert_mask(img):
    return 1 - img


def mask_image(image, mask, value: int = 0):
    image[mask] = value
    return image



"""
This code was a direct implemetation for the work describes in the paper

@inproceedings{gupta2022learning,
  title={Learning Topological Interactions for Multi-Class Medical Image Segmentation},
  author={Gupta, Saumya and Hu, Xiaoling and Kaan, James and Jin, Michael and Mpoy, Mutshipay and Chung, Katherine and Singh, Gagandeep and Saltz, Mary and Kurc, Tahsin and Saltz, Joel and others},
  booktitle={Computer Vision--ECCV 2022: 17th European Conference, Tel Aviv, Israel, October 23--27, 2022, Proceedings, Part XXIX},
  pages={701--718},
  year={2022},
  organization={Springer}
}

https://www.ecva.net/papers/eccv_2022/papers_ECCV/papers/136890691.pdf

and provided by the github repo:  https://github.com/TopoXLab/TopoInteraction
"""

import numpy as np
import torch

"""
The proposed topological interaction (TI) module encodes topological interactions by computing the critical voxels map. The critical voxels map contains the locations which induce errors in the topological interactions. The TI loss is introduced based on the topological interaction module.
"""


class TI_Loss(torch.nn.Module):
    def __init__(self, dim, connectivity, inclusion, exclusion, min_thick=1):
        """
        :param dim: 2 if 2D; 3 if 3D
        :param connectivity: 4 or 8 for 2D; 6 or 26 for 3D
        :param inclusion: list of [A,B] classes where A is completely surrounded by B.
        :param exclusion: list of [A,C] classes where A and C exclude each other.
        :param min_thick: Minimum thickness/separation between the two classes. Only used if connectivity is 8 for 2D or 26 for 3D
        """
        super(TI_Loss, self).__init__()

        self.dim = dim
        self.connectivity = connectivity
        self.min_thick = min_thick
        self.interaction_list = []
        self.sum_dim_list = None
        self.conv_op = None
        self.apply_nonlin = lambda x: torch.nn.functional.softmax(x, 1)
        self.ce_loss_func = torch.nn.CrossEntropyLoss(reduction="none")

        if self.dim == 2:
            self.sum_dim_list = [1, 2, 3]
            self.conv_op = torch.nn.functional.conv2d
        elif self.dim == 3:
            self.sum_dim_list = [1, 2, 3, 4]
            self.conv_op = torch.nn.functional.conv3d

        self.set_kernel()

        for inc in inclusion:
            temp_pair = []
            temp_pair.append(True)  # type inclusion
            temp_pair.append(inc[0])
            temp_pair.append(inc[1])
            self.interaction_list.append(temp_pair)

        for exc in exclusion:
            temp_pair = []
            temp_pair.append(False)  # type exclusion
            temp_pair.append(exc[0])
            temp_pair.append(exc[1])
            self.interaction_list.append(temp_pair)

    def set_kernel(self):
        """
        Sets the connectivity kernel based on user's sepcification of dim, connectivity, min_thick
        """
        k = 2 * self.min_thick + 1
        if self.dim == 2:
            if self.connectivity == 4:
                np_kernel = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]])
            elif self.connectivity == 8:
                np_kernel = np.ones((k, k))

        elif self.dim == 3:
            if self.connectivity == 6:
                np_kernel = np.array(
                    [
                        [[0, 0, 0], [0, 1, 0], [0, 0, 0]],
                        [[0, 1, 0], [1, 1, 1], [0, 1, 0]],
                        [[0, 0, 0], [0, 1, 0], [0, 0, 0]],
                    ]
                )
            elif self.connectivity == 26:
                np_kernel = np.ones((k, k, k))

        self.kernel = torch.from_numpy(
            np.expand_dims(np.expand_dims(np_kernel, axis=0), axis=0)
        )

    def topological_interaction_module(self, P):
        """
        Given a discrete segmentation map and the intended topological interactions, this module computes the critical voxels map.
        :param P: Discrete segmentation map
        :return: Critical voxels map
        """
        critical_voxels_map = torch.zeros_like(P, dtype=torch.double)

        for ind, interaction in enumerate(self.interaction_list):
            interaction_type = interaction[0]
            label_A = interaction[1]
            label_C = interaction[2]

            # Get Masks
            mask_A = torch.where(P == label_A, 1.0, 0.0).double()
            if interaction_type:
                mask_C = torch.where(P == label_C, 1.0, 0.0).double()
                mask_C = torch.logical_or(mask_C, mask_A).double()
                mask_C = torch.logical_not(mask_C).double()
            else:
                mask_C = torch.where(P == label_C, 1.0, 0.0).double()

            # Get Neighbourhood Information
            neighbourhood_C = self.conv_op(mask_C, self.kernel.double(), padding="same")
            neighbourhood_C = torch.where(neighbourhood_C >= 1.0, 1.0, 0.0)
            neighbourhood_A = self.conv_op(mask_A, self.kernel.double(), padding="same")
            neighbourhood_A = torch.where(neighbourhood_A >= 1.0, 1.0, 0.0)

            # Get the pixels which induce errors
            violating_A = neighbourhood_C * mask_A
            violating_C = neighbourhood_A * mask_C
            violating = violating_A + violating_C
            violating = torch.where(violating >= 1.0, 1.0, 0.0)

            critical_voxels_map = torch.logical_or(
                critical_voxels_map, violating
            ).double()

        return critical_voxels_map

    def forward(self, x, y):
        """
        The forward function computes the TI loss value.
        :param x: Likelihood map of shape: b, c, x, y(, z) with c = total number of classes
        :param y: GT of shape: b, c, x, y(, z) with c=1. The GT should only contain values in [0,L) range where L is the total number of classes.
        :return:  TI loss value
        """

        if x.device.type == "cuda":
            self.kernel = self.kernel.cuda(x.device.index)

        # Obtain discrete segmentation map
        x_softmax = self.apply_nonlin(x)
        P = torch.argmax(x_softmax, dim=1)
        P = torch.unsqueeze(P.double(), dim=1)
        del x_softmax

        # Call the Topological Interaction Module
        critical_voxels_map = self.topological_interaction_module(P)

        # Compute the TI loss value
        ce_tensor = torch.unsqueeze(
            self.ce_loss_func(x.double(), y[:, 0].long()), dim=1
        )
        ce_tensor[:, 0] = ce_tensor[:, 0] * torch.squeeze(critical_voxels_map, dim=1)
        ce_loss_value = ce_tensor.sum(dim=self.sum_dim_list).mean()

        return ce_loss_value



# ADAPTED FROM https://pytorch.org/vision/main/_modules/torchvision/models/segmentation/deeplabv3.html  # noqa E501
# and https://discuss.pytorch.org/t/how-to-modify-deeplabv3-and-fcn-models-for-grayscale-images/52688  # noqa E501

import torch
from mmv_im2im.utils.misc import parse_config_func_without_params


class Net(torch.nn.Module):
    def __init__(
        self,
        backbone,
        pretrained: bool = False,
        pretrained_backbone: bool = True,
        in_channels: int = 3,
        num_classes: int = 21,
        aux_loss: bool = None,
    ):
        super().__init__()
        params = {
            "progress": False,
            "num_classes": num_classes,
            "aux_loss": aux_loss,
            "pretrained": pretrained,
            "pretrained_backbone": pretrained_backbone,
        }
        info = {"module_name": "torchvision.models.segmentation", "params": params}
        if backbone == "deeplabv3_resnet50":
            info["func_name"] = "deeplabv3_resnet50"
            my_func = parse_config_func_without_params(info)
            self.net = my_func(**info["params"])
        elif backbone == "deeplabv3_resnet101":
            info["func_name"] = "deeplabv3_resnet101"
            my_func = parse_config_func_without_params(info)
            self.net = my_func(**info["params"])
        elif backbone == "deeplabv3_mobilenet_v3_large":
            info["func_name"] = "deeplabv3_mobilenet_v3_large"
            my_func = parse_config_func_without_params(info)
            self.net = my_func(**info["params"])
        if in_channels != 3:
            self.net.backbone.conv1 = torch.nn.Conv2d(
                in_channels, 64, kernel_size=7, stride=2, padding=3, bias=False
            )

    def forward(self, x):
        y_hat = self.net(x)
        return y_hat["out"]



# -*- coding: utf-8 -*-

"""Top-level package for MMV Im2Im Transformation."""

__author__ = "Jianxu Chen"
__email__ = "jianxuchen.ai@gmail.com"
# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "0.6.2"


def get_module_version():
    return __version__


from .proj_tester import ProjectTester  # noqa F401
from .proj_trainer import ProjectTrainer  # noqa F401



import numpy as np


def mean_prediction(samples):
    stacked_predictions = np.array(samples)
    mean_volume = np.mean(stacked_predictions, axis=0)

    return mean_volume


def max_prediction(samples):
    stacked_predictions = np.array(samples)
    max_volume = np.max(stacked_predictions, axis=0)

    return max_volume


def variance_prediction(samples):
    stacked_predictions = np.array(samples)
    variance_volume = np.var(stacked_predictions, axis=0)
    return variance_volume


def add_prediction(samples):
    stacked_predictions = np.array(samples)
    sum_volume = np.sum(stacked_predictions, axis=0)
    return sum_volume



from typing import Union
from importlib import import_module
import numpy as np
from skimage.measure import regionprops
from skimage.segmentation import relabel_sequential
from skimage.morphology import binary_opening
import torch


def extract_segmentation(
    im: Union[np.ndarray, torch.Tensor],
    channel: int,
    cutoff: Union[float, str] = None,
    batch_dim: bool = True,
) -> np.ndarray:
    """extract segmentation from a prediction

    Parameters:
    -------------
    im: ndarray or torch.Tensor
        the multi-class prediction (1, C, W, H) or (1, C, Z, Y, X)
    channel: int
        which channel to select
    cutoff: float or str
        either a fixed cutoff value or a segmentation method from skimage, default is
        None (do not apply any cutoff)
    batch_dim: bool
        whether there is a batch dimension (default is True)
    """

    # convert tensor to numpy
    if torch.is_tensor(im):
        im = im.cpu().numpy()
    if batch_dim:
        assert (
            len(im.shape) == 4 or len(im.shape) == 5
        ), "extract seg with batch_dim only accepts 4D/5D"
        assert im.shape[0] == 1, "extract seg with batch_dim requires first dim to be 1"
        prob = im[0, channel, :]
    else:
        assert (
            len(im.shape) == 3 or len(im.shape) == 4
        ), "extract seg without batch_dim only accepts 3D/4D"
        prob = im[channel, :]

    if cutoff is None:
        return prob
    else:
        if isinstance(cutoff, str):
            th_module = import_module("skimage.filters")
            try:
                th_func = getattr(th_module, "threshold_" + cutoff.lower())
            except Exception:
                raise ValueError("unsupported threhsold method")
            cutoff = th_func(prob)
        elif isinstance(cutoff, float):
            # just to confirm the type
            pass
        else:
            raise NotImplementedError("cutoff method only str or float")

        seg = prob > cutoff
        seg = seg.astype(np.uint8)
        seg[seg > 0] = 255

        return seg


def generate_classmap(im: Union[np.ndarray, torch.Tensor]) -> np.ndarray:
    """generate the segmentation classmap from model prediction

    Parameters:
    -------------
    im: ndarray or torch.Tensor
        the multi-class prediction (1, C, W, H) or (1, C, Z, Y, X)
    """

    # convert tensor to numpy
    if torch.is_tensor(im):
        im = im.cpu().numpy()
    assert len(im.shape) == 4 or len(im.shape) == 5, "extract seg only accepts 4D/5D"
    assert im.shape[0] == 1, "extract seg requires first dim to be 1"

    classmap = np.argmax(im, axis=1).astype(np.uint8)
    return classmap


def prune_labels_by_size(labeled_image, min_size=10):
    # Iterate through labeled objects and calculate their areas
    for region in regionprops(labeled_image):
        if region.num_pixels < min_size:
            labeled_image[labeled_image == region.label] = 0
    filtered_labeled_image, _, _ = relabel_sequential(labeled_image)
    return filtered_labeled_image


def remove_isolated_pixels(labeled_image):
    for region in regionprops(labeled_image):
        single_obj = labeled_image == region.label
        single_obj_clean = binary_opening(single_obj)
        labeled_image[labeled_image == region.label] = 0
        if np.any(single_obj_clean):
            labeled_image[single_obj_clean] = region.label

    labeled_image_clean, _, _ = relabel_sequential(labeled_image)

    return labeled_image_clean



import functools
import torch.nn as nn

from monai.networks.blocks import ResidualUnit, Convolution
from monai.networks.nets import UNet

from mmv_im2im.utils.misc import (
    parse_config_func_without_params,
    parse_config_func,
    parse_config,
)


class preset_generator_resent(nn.Module):
    def __init__(
        self,
        spatial_dims,
        in_channels,
        out_channels,
        n_down_blocks,
        n_res_blocks,
        nf=64,
        norm_layer="INSTANCE",
        use_dropout=None,
    ):
        super().__init__()
        if spatial_dims == 2:
            first_layer = functools.partial(nn.ReflectionPad2d)
        elif spatial_dims == 3:
            first_layer = functools.partial(nn.ReplicationPad3d)
        else:
            raise NotImplementedError("only 2d and 3d data are supported")

        model = [
            first_layer(3),
            Convolution(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=nf,
                norm=norm_layer,
                kernel_size=7,
                act="RELU",
                padding=0,
            ),
        ]

        # add downsampling layers
        for i in range(n_down_blocks):
            mult = 2**i
            model += [
                Convolution(
                    spatial_dims=spatial_dims,
                    in_channels=nf * mult,
                    out_channels=nf * mult * 2,
                    kernel_size=3,
                    strides=2,
                    padding=1,
                    norm=norm_layer,
                    act="RELU",
                )
            ]

        # add ResNet blocks
        mult = 2**n_down_blocks
        for i in range(n_res_blocks):
            model += [
                ResidualUnit(
                    spatial_dims=spatial_dims,
                    in_channels=nf * mult,
                    out_channels=nf * mult,
                    kernel_size=3,
                    strides=1,
                    norm=norm_layer,
                    dropout=use_dropout,
                    act="RELU",
                )
            ]

        # add upsampling blocks
        for i in range(n_down_blocks):
            mult = 2 ** (n_down_blocks - i)
            model += [
                Convolution(
                    spatial_dims=spatial_dims,
                    in_channels=nf * mult,
                    out_channels=int(nf * mult / 2),
                    kernel_size=3,
                    strides=2,
                    padding=1,
                    output_padding=1,
                    norm=norm_layer,
                    act="RELU",
                    is_transposed=True,
                )
            ]

        # add final layer
        model += [
            first_layer(3),
            Convolution(
                spatial_dims=spatial_dims,
                in_channels=nf,
                out_channels=out_channels,
                kernel_size=7,
                conv_only=True,
                padding=0,
            ),
        ]

        self.model = nn.Sequential(*model)

    def forward(self, x):
        return self.model(x)


class generator_encoder_decoder(nn.Module):
    def __init__(self, model_info):
        super().__init__()
        # initial block
        init_block = parse_config_func(model_info["init_block"])
        model = [init_block]

        # downsampling blocks
        down_block = parse_config_func_without_params(model_info["down_block"])
        down_channels = model_info["down_block"]["params"].pop("channels")
        prev_channel = model_info["init_block"]["params"]["out_channel"]
        for this_channel in down_channels:
            model.append(
                down_block(
                    in_channels=prev_channel,
                    out_channels=this_channel,
                    **model_info["down_block"]["params"],
                )
            )

        # residual block
        res_block = parse_config_func_without_params(model_info["res_block"])
        prev_channel = down_channels[-1]
        res_channels = model_info["res_block"]["params"].pop("channels")
        for this_channel in res_channels:
            model.append(
                res_block(
                    in_channels=prev_channel,
                    out_channels=this_channel,
                    **model_info["res_block"]["params"],
                )
            )

        # up blocks
        up_block = parse_config_func_without_params(model_info["up_block"])
        prev_channel = res_channels[-1]
        up_channels = model_info["up_block"]["params"].pop("channels")
        for this_channel in up_channels:
            model.append(
                up_block(
                    in_channels=prev_channel,
                    out_channels=this_channel,
                    **model_info["up_block"]["params"],
                )
            )

        # final block
        final_block = parse_config_func(model_info["final_block"])
        model.append(final_block)

        self.model = nn.Sequential(*model)

    def forward(self, x):
        return self.model(x)


def define_generator(model_info):
    if "type" in model_info:
        if model_info["type"] == "predefined_resnet":
            net = preset_generator_resent(**model_info["params"])
        elif model_info["type"] == "predefined_unet":
            net = UNet(**model_info["params"])
        elif model_info["type"] == "generic_encoder_decoder":
            net = generator_encoder_decoder(**model_info)
    else:
        net = parse_config(model_info)

    return net


class patch_discriminator(nn.Module):
    def __init__(
        self,
        spatial_dims,
        in_channels,
        nf,
        n_layers,
        norm_layer,
        return_features: bool = False,
    ):
        super().__init__()

        blocks = [
            Convolution(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=nf,
                kernel_size=4,
                strides=2,
                padding=1,
                conv_only=True,
            ),  # no norm for first layer
            nn.LeakyReLU(0.2, True),
        ]

        nf_mult = 1
        nf_mult_prev = 1
        for n in range(1, n_layers):  # gradually increase the number of filters
            nf_mult_prev = nf_mult
            nf_mult = min(2**n, 8)
            blocks += [
                Convolution(
                    spatial_dims=spatial_dims,
                    in_channels=nf * nf_mult_prev,
                    out_channels=nf * nf_mult,
                    kernel_size=4,
                    strides=2,
                    padding=1,
                    norm=norm_layer,
                    act=("leakyrelu", {"negative_slope": 0.2, "inplace": True}),
                )
            ]

        blocks += [
            Convolution(
                spatial_dims=spatial_dims,
                in_channels=nf * nf_mult,
                out_channels=1,
                kernel_size=4,
                strides=1,
                padding=1,
                conv_only=True,
            )
        ]

        self.return_features = return_features
        if return_features:
            self.n_blocks = len(blocks)
            for i in range(self.n_blocks):
                setattr(self, "block_{}".format(i), nn.Sequential(*blocks[i]))
        else:
            self.model = nn.Sequential(*blocks)

    def forward(self, x):
        if self.return_features:
            result = [x]
            for i in range(self.n_blocks):
                block = getattr(self, "block_{}".format(i))
                result.append(block(result[-1]))

            return result[1:]  # except for the input
        else:
            return self.model(x)


class multiscale_discriminator(nn.Module):
    def __init__(self, num_discriminator, **kwargs):
        super().__init__()

        for i in range(num_discriminator):
            setattr(self, "Scale_{}".format(str(i)), patch_discriminator(**kwargs))
        self.n_D = num_discriminator
        self.spatial_dimenstion = kwargs["spatial_dims"]

    def forward(self, x):
        result = []
        for i in range(self.n_D):
            result.append(getattr(self, "Scale_{}".format(i))(x))
            if i != self.n_D - 1:
                if self.spatial_dimenstion == 2:
                    x = nn.AvgPool2d(
                        kernel_size=3, padding=1, stride=2, count_include_pad=False
                    )(x)
                elif self.spatial_dimenstion == 3:
                    x = nn.AvgPool3d(
                        kernel_size=3, padding=1, stride=2, count_include_pad=False
                    )(x)
        return result


def define_discriminator(model_info):
    if model_info["type"] == "predefined_basic":
        net = patch_discriminator(**model_info["params"])
    elif model_info["type"] == "predefined_multiscale":
        net = multiscale_discriminator(**model_info["params"])
    else:
        raise NotImplementedError("only predefined discriminator is supported")

    return net



from dataclasses import dataclass
from pathlib import Path
from pyrallis import field

import argparse
import dataclasses
import sys
import warnings
from argparse import HelpFormatter, Namespace
from collections import defaultdict
from logging import getLogger
from typing import Dict, List, Sequence, Text, Type, Union, TypeVar, Generic, Optional

from pyrallis import utils, cfgparsing
from pyrallis.help_formatter import SimpleHelpFormatter
from pyrallis.parsers import decoding
from pyrallis.utils import Dataclass, PyrallisException
from pyrallis.wrappers import DataclassWrapper

logger = getLogger(__name__)

T = TypeVar("T")


class ArgumentParser(Generic[T], argparse.ArgumentParser):
    def __init__(
        self,
        config_class: Type[T],
        config: Optional[str] = None,
        formatter_class: Type[HelpFormatter] = SimpleHelpFormatter,
        *args,
        **kwargs,
    ):
        """Creates an ArgumentParser instance."""
        kwargs["formatter_class"] = formatter_class
        super().__init__(*args, **kwargs)

        # constructor arguments for the dataclass instances.
        # (a Dict[dest, [attribute, value]])
        self.constructor_arguments: Dict[str, Dict] = defaultdict(dict)

        self._wrappers: List[DataclassWrapper] = []

        self.config = config
        self.config_class = config_class

        self._assert_no_conflicts()
        self.add_argument(
            f"--{utils.CONFIG_ARG}",
            type=str,
            help="Path for a config file to parse with pyrallis",
        )
        self.set_dataclass(config_class)

    def set_dataclass(
        self,
        dataclass: Union[Type[Dataclass], Dataclass],
        prefix: str = "",
        default: Union[Dataclass, Dict] = None,
        dataclass_wrapper_class: Type[DataclassWrapper] = DataclassWrapper,
    ):
        """Adds command-line arguments for the fields of `dataclass`."""
        if not isinstance(dataclass, type):
            default = dataclass if default is None else default
            dataclass = type(dataclass)

        new_wrapper = dataclass_wrapper_class(dataclass, prefix=prefix, default=default)
        self._wrappers.append(new_wrapper)
        self._wrappers += new_wrapper.descendants

        for wrapper in self._wrappers:
            logger.debug(
                f"Adding arguments for dataclass: {wrapper.dataclass} "
                f"at destination {wrapper.dest}"
            )
            wrapper.add_arguments(parser=self)

    def _assert_no_conflicts(self):
        """Checks for a field name that conflicts with utils.CONFIG_ARG"""
        if utils.CONFIG_ARG in [
            field.name for field in dataclasses.fields(self.config_class)
        ]:
            raise PyrallisException(
                f"{utils.CONFIG_ARG} is a reserved word for pyrallis"
            )

    def parse_args(self, args=None, namespace=None) -> T:
        return super().parse_args(args, namespace)

    def parse_known_args(
        self,
        args: Sequence[Text] = None,
        namespace: Namespace = None,
        attempt_to_reorder: bool = False,
    ):
        # NOTE: since the usual ArgumentParser.parse_args() calls
        # parse_known_args, we therefore just need to overload the
        # parse_known_args method to support both.
        if args is None:
            # args default to the system args
            args = sys.argv[1:]
        else:
            # make sure that args are mutable
            args = list(args)

        if "--help" not in args:
            for action in self._actions:
                # TODO: Find a better way to do that?
                action.default = (
                    argparse.SUPPRESS
                )  # To avoid setting of defaults in actual run
                action.type = (
                    str  # In practice, we want all processing to happen with yaml
                )
        parsed_args, unparsed_args = super().parse_known_args(args, namespace)

        parsed_args = self._postprocessing(parsed_args)
        return parsed_args, unparsed_args

    def print_help(self, file=None):
        return super().print_help(file)

    def _postprocessing(self, parsed_args: Namespace) -> T:
        logger.debug("\nPOST PROCESSING\n")
        logger.debug(f"(raw) parsed args: {parsed_args}")

        parsed_arg_values = vars(parsed_args)

        for key in parsed_arg_values:
            parsed_arg_values[key] = cfgparsing.parse_string(parsed_arg_values[key])

        config = self.config  # Could be NONE

        if utils.CONFIG_ARG in parsed_arg_values:
            new_config = parsed_arg_values[utils.CONFIG_ARG]
            if config is not None:
                warnings.warn(
                    UserWarning(f"Overriding default {config} with {new_config}")
                )
            ######################################################################
            # adapted from original implementation in pyrallis
            ######################################################################
            if Path(new_config).is_file():
                # pass in a absolute path
                config = new_config
            else:
                new_config = str(new_config)
                print(f"trying to locate preset config for {new_config} ...")

                config = Path(__file__).parent / f"preset_{new_config}.yaml"
            del parsed_arg_values[utils.CONFIG_ARG]

        if config is not None:
            print(f"loading configuration from {config} ...")
            file_args = cfgparsing.load_config(open(config, "r"))
            file_args = utils.flatten(file_args, sep=".")
            file_args.update(parsed_arg_values)
            parsed_arg_values = file_args
            print("configuration loading is completed")

        deflat_d = utils.deflatten(parsed_arg_values, sep=".")
        cfg = decoding.decode(self.config_class, deflat_d)

        return cfg


def parse_adaptor(
    config_class: Type[T],
    config: Optional[Union[Path, str]] = None,
    args: Optional[Sequence[str]] = None,
) -> T:
    parser = ArgumentParser(config_class=config_class, config=config)
    return parser.parse_args(args)


@dataclass
class DataloaderModuleConfig:
    """low-level config for data: specifically about the dataloader"""

    # which dataset module to use: default is PersistentDataset
    dataloader_type: Dict = field(
        default={"module_name": "monai.data", "func_name": "PersistentDataset"},
        is_mutable=True,
    )

    # the parameters to be passed in when making the dataset (e.g. PersistentDataset)
    # find full API in the corresponding Dataset function
    # e.g. https://docs.monai.io/en/stable/data.html#persistentdataset
    dataset_params: Dict = field(
        default={"cache_dir": "./tmp", "pickle_protocol": 5}, is_mutable=True
    )

    # the parameters to be passed in when making the DataLoader
    # find full API here:
    # https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader
    dataloader_params: Dict = field(
        default={
            "batch_size": 1,
            "pin_memory": True,
            "num_workers": 2,
        },
        is_mutable=True,
    )

    # whether to load part of the dataset, and periodically reload
    partial_loader: float = field(default=1.0)


@dataclass
class DataloaderConfig:
    """Mid-level config for data: about dataloader module"""

    # The ratio of train/validation split
    train_val_ratio: float = field(default=0.1)

    # config for the training dataloader (see DataloaderModuleConfig)
    train: DataloaderModuleConfig = field(default_factory=DataloaderModuleConfig)

    # config for the validation dataloader (see DataloaderModuleConfig)
    val: DataloaderModuleConfig = field(default_factory=DataloaderModuleConfig)


@dataclass
class InferOutConfig:
    """config for the output of inference"""

    # The path to save output
    path: Union[Path, str, Dict] = field(default=None)

    # the suffix to attach in the filename
    suffix: str = field(default="_pred")


@dataclass
class InferInConfig:
    """config for the input to inference"""

    # The path to apply the model
    dir: Union[Path, str, Dict] = field(default=None)

    # the type of data to load
    data_type: str = field(default="tiff")

    # the parameters for reader
    reader_params: Dict = field(default=None)


@dataclass
class DataConfig:
    """Top-level config for data: major components"""

    # The type of data: "pair" | "unpair" | "embedseg"
    category: str = field(default=None)

    # The data path
    data_path: Union[Path, str, Dict] = field(default=None)

    # save pre-processed data into a cache folder (currently, only for embedseg)
    cache_path: Union[Path, str] = field(default=None)

    # about the inference output (Inference Only)
    inference_output: InferOutConfig = field(default_factory=InferOutConfig)

    # about the inference input (Inference Only)
    inference_input: InferInConfig = field(default_factory=InferInConfig)

    # config for dataloader (see DataloaderConfig)
    dataloader: DataloaderConfig = field(default_factory=DataloaderConfig)

    # what to do in pre-processing (see examples in preset configs)
    preprocess: List[Dict] = field(default=None)

    # what to do in postprocessing (see examples in preset configs)
    postprocess: List[Dict] = field(default=None)

    # what to do in data augmentation (see examples in preset configs)
    augmentation: List[Dict] = field(default=None)

    # global variable that can be used to verify or overwrite other related settings
    patch_size: List = field(default=None, is_mutable=True)

    # extra parameters for specific methods:
    extra: Dict = field(default=None, is_mutable=True)


@dataclass
class ModelConfig:
    """Config for model"""

    # the type of model: FCN | pix2pix | cyclegan | embedseg
    framework: str = field(default=None)

    # the exact network configuration (see preset configs for example)
    net: Dict = field(default=None, is_mutable=True)

    # the config for criterion (see preset configs for example)
    criterion: Dict = field(default=None, is_mutable=True)

    # the config for optimizer (see preset configs for example)
    optimizer: Dict = field(default=None, is_mutable=True)

    # the config for learning scheduler (see preset configs for example)
    scheduler: Dict = field(default=None, is_mutable=True)

    # the checkpoint to load (for inference)
    checkpoint: Union[Path, str] = field(default=None)

    # extra for special parameters of specific method
    model_extra: Dict = field(default=None, is_mutable=True)


@dataclass
class TrainerConfig:
    """Config for how to run the training"""

    # whether to save sample outputs at the beginning of each epoch
    verbose: bool = field(default=False)

    # the parameters to be passed into pytorch-lightning trainer
    params: Dict = field(default=None)

    # the config for callbacks
    callbacks: List[Dict] = field(default=None, is_mutable=True)

    # global variable that can be used to overwrite gpus in trainer
    gpus: Union[int, List[int]] = field(default=None, is_mutable=True)


@dataclass
class ProgramConfig:
    """the main configuration for the whole program"""

    # mode: "train" | "infernece" (#TODO future version will add "evaluation")
    mode: str = field(default=None)

    # the configuration for data (path, dataloader, transform, etc.)
    data: DataConfig = field(default_factory=DataConfig)

    # the configuration for model (e.g., network, loss, etc.)
    model: ModelConfig = field(default_factory=ModelConfig)

    # the configuration for trainer
    trainer: TrainerConfig = field(default_factory=TrainerConfig)


def configuration_validation(cfg):
    # check 1: partial_loader should be a value between 0 (excluded) and 1 (included)
    assert (
        cfg.data.dataloader.train.partial_loader <= 1.0
        and cfg.data.dataloader.train.partial_loader > 0
    ), "partial loading percentage for training loader is not valid"

    # check 2: if partial loading is used, make sure reloading is enabled in trainer
    if cfg.data.dataloader.train.partial_loader < 1.0:
        if "reload_dataloaders_every_n_epochs" not in cfg.trainer.params:
            cfg.trainer.params["reload_dataloaders_every_n_epochs"] = 5

    # check 3: partial_loader for validation dataloader should be 1.0
    if cfg.data.dataloader.val.partial_loader != 1.0:
        cfg.data.dataloader.val.partial_loader = 1.0
        warnings.warn(
            UserWarning("partial loading for validation step is not allowed.")
        )

    # check 4: for embedseg, patch_size needs to be consistent with grid in criterion
    if cfg.data.patch_size is not None:
        grid = cfg.model.criterion["params"]["grid_x"]
        assert (
            grid == cfg.data.patch_size[-1]
        ), f"grid_x is set as {grid}, needs to be the same as the dimX of patch_size {cfg.data.patch_size}"  # noqa E501

        grid = cfg.model.criterion["params"]["grid_y"]
        assert (
            grid == cfg.data.patch_size[-2]
        ), f"grid_y is set as {grid}, needs to be the same as the dimY of patch_size {cfg.data.patch_size}"  # noqa E501

        if len(cfg.data.patch_size) == 3:
            grid = cfg.model.criterion["params"]["grid_z"]
            assert (
                grid == cfg.data.patch_size[0]
            ), f"grid_z is set as {grid}, needs to be the same as the dimZ of patch_size {cfg.data.patch_size}"  # noqa E501

    # check 5, if a global GPU number is set, update the value in trainer
    if cfg.trainer.gpus is not None:
        cfg.trainer.params["gpus"] = cfg.trainer.gpu

    # check 5, if PersistentDataset is used, make sure add a tmpdir in subdirectory
    # (otherwise may cause hash errors)
    if cfg.mode == "train":
        if (
            cfg.data.dataloader.train.dataloader_type["func_name"]
            == "PersistentDataset"
        ):
            assert (
                cfg.data.dataloader.val.dataloader_type["func_name"]
                == "PersistentDataset"
            ), "currently, train and val can only use persisten loader together"

        if cfg.data.dataloader.val.dataloader_type["func_name"] == "PersistentDataset":
            assert (
                cfg.data.dataloader.train.dataloader_type["func_name"]
                == "PersistentDataset"
            ), "currently, train and val can only use persisten loader together"
            if (
                cfg.data.dataloader.train.dataset_params["cache_dir"]
                != cfg.data.dataloader.val.dataset_params["cache_dir"]
            ):
                cfg.data.dataloader.val.dataset_params["cache_dir"] = (
                    cfg.data.dataloader.train.dataset_params["cache_dir"]
                )
                warnings.warn(
                    UserWarning(
                        "The cache dir of PersistentDataset for validation was"
                        "overwritten to the cache dir for train. Currently, they"
                        "have to be the same to allow easy clean up"
                    )
                )

    return cfg



import torch
import torch.nn as nn
import mmv_im2im.models.nets.erfnet_3d as erfnet


class BranchedERFNet_3d(nn.Module):
    def __init__(self, num_classes, input_channels=1, encoder=None):
        super().__init__()

        print("Creating branched erfnet 3d with {} classes".format(num_classes))
        if encoder is None:
            self.encoder = erfnet.Encoder(sum(num_classes), input_channels)
        else:
            self.encoder = encoder

        self.decoders = nn.ModuleList()
        for n in num_classes:
            self.decoders.append(erfnet.Decoder(n))

    def init_output(self, n_sigma=1):
        with torch.no_grad():
            output_conv = self.decoders[0].output_conv
            print("initialize last layer with size: ", output_conv.weight.size())

            output_conv.weight[:, 0:3, :, :, :].fill_(0)
            output_conv.bias[0:3].fill_(0)

            output_conv.weight[:, 3 : 3 + n_sigma, :, :, :].fill_(0)
            output_conv.bias[3 : 3 + n_sigma].fill_(1)

    def forward(self, input, only_encode=False):
        if only_encode:
            return self.encoder.forward(input, predict=True)
        else:
            output = self.encoder(input)

        return torch.cat([decoder.forward(output) for decoder in self.decoders], 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from typing import Union
from dask.array.core import Array as DaskArray
from numpy import ndarray as NumpyArray
from importlib import import_module
from pathlib import Path
import numpy as np
from bioio import BioImage
from bioio.writers import OmeTiffWriter
import torch
from mmv_im2im.utils.misc import parse_config
from mmv_im2im.utils.for_transform import parse_monai_ops_vanilla
from skimage.io import imsave as save_rgb
import bioio_tifffile
from mmv_im2im.utils.urcentainity_extractor import (
    Hole_Correction,
    Thickness_Corretion,
    Remove_objects,
    Extract_Uncertainty_Maps,
    perturb_image,
    Perycites_correction,
)
from monai.inferers import sliding_window_inference
import itertools
from mmv_im2im.utils.multi_pred import (
    variance_prediction,
    mean_prediction,
    max_prediction,
    add_prediction,
)
from bioio_base.types import PhysicalPixelSizes

# https://pytorch-lightning.readthedocs.io/en/latest/starter/introduction_guide.html#predicting
###############################################################################

log = logging.getLogger(__name__)

###############################################################################


class MapExtractor(object):
    """
    entry for training models

    Parameters
    ----------
    cfg: configuration
    """

    def __init__(self, cfg):
        # extract the three major chuck of the config
        self.model_cfg = cfg.model
        self.data_cfg = cfg.data

        # define variables
        self.model = None
        self.data = None
        self.pre_process = None
        self.cpu = False
        self.spatial_dims = -1

    def setup_model(self):
        model_category = self.model_cfg.framework
        model_module = import_module(f"mmv_im2im.models.pl_{model_category}")
        my_model_func = getattr(model_module, "Model")
        self.model = my_model_func(self.model_cfg, train=False)

        if (
            self.model_cfg.model_extra is not None
            and "cpu_only" in self.model_cfg.model_extra
            and self.model_cfg.model_extra["cpu_only"]
        ):
            self.cpu = True
            checkpoint = torch.load(
                self.model_cfg.checkpoint,
                map_location=torch.device("cpu"),
                weights_only=False,
            )
        else:
            checkpoint = torch.load(self.model_cfg.checkpoint, weights_only=False)

        if isinstance(checkpoint, dict) and "state_dict" in checkpoint:

            pre_train = checkpoint
            pre_train["state_dict"].pop("criterion.xym", None)
            pre_train["state_dict"].pop("criterion.xyzm", None)
            self.model.load_state_dict(pre_train["state_dict"], strict=False)
        else:

            state_dict = checkpoint
            state_dict.pop("criterion.xym", None)
            state_dict.pop("criterion.xyzm", None)
            self.model.load_state_dict(state_dict, strict=False)

        if not self.cpu:
            self.model.cuda()

        self.model.eval()

    def setup_data_processing(self):
        # determine spatial dimension from reader parameters
        if "Z" in self.data_cfg.inference_input.reader_params["dimension_order_out"]:
            self.spatial_dims = 3
        else:
            self.spatial_dims = 2

        # prepare data preprocessing if needed
        if self.data_cfg.preprocess is not None:
            # load preprocessing transformation
            self.pre_process = parse_monai_ops_vanilla(self.data_cfg.preprocess)

    def process_one_image(
        self, img: Union[DaskArray, NumpyArray], out_fn: Union[str, Path] = None
    ):

        if isinstance(img, DaskArray):
            # Perform the prediction
            x = img.compute()

        elif isinstance(img, NumpyArray):
            x = img
        else:
            raise ValueError("invalid image")

        # check if need to add channel dimension
        if len(x.shape) == self.spatial_dims:
            x = np.expand_dims(x, axis=0)

        # convert the numpy array to float tensor
        x = torch.tensor(x.astype(np.float32))

        # run pre-processing on tensor if needed

        if self.pre_process is not None:
            x = self.pre_process(x)
            x = x[0]

        # choose different inference function for different types of models
        # the input here is assumed to be a tensor
        with torch.no_grad():
            # add batch dimension and move to GPU

            if self.cpu:
                x = torch.unsqueeze(x, dim=0)
            else:
                x = torch.unsqueeze(x, dim=0).cuda()

            # TODO: add convert to tensor with proper type, similar to torchio check

            if (
                self.model_cfg.model_extra is not None
                and "sliding_window_params" in self.model_cfg.model_extra
            ):
                y_hat = sliding_window_inference(
                    inputs=x,
                    predictor=self.model,
                    device=torch.device("cpu"),
                    **self.model_cfg.model_extra["sliding_window_params"],
                )

                # currently, we keep sliding window stiching step on CPU, but assume
                # the output is on GPU (see note below). So, we manually move the data
                # back to GPU
                if not self.cpu:
                    y_hat = y_hat.cuda()
            else:
                y_hat = self.model(x)

        ###############################################################################
        #
        # Note: currently, we assume y_hat is still on gpu, because embedseg clustering
        # step is still only running on GPU (possible on CPU, need to some update on
        # grid loading). All the post-procesisng functions we tested so far can accept
        # tensor on GPU. If it is from mmv_im2im.post_processing, it will automatically
        # convert the tensor to a numpy array and return the result as numpy array; if
        # it is from monai.transforms, it is tensor in and tensor out. We have two items
        # as #TODO: (1) we will extend post-processing functions in mmv_im2im to work
        # similarly to monai transforms, ie. ndarray in ndarray out or tensor in tensor
        # out. (2) allow yaml config to control if we want to run post-processing on
        # GPU tensors or ndarrays
        #
        ##############################################################################

        # do post-processing on the prediction
        if self.data_cfg.postprocess is not None:
            pp_data = y_hat
            for pp_info in self.data_cfg.postprocess:
                pp = parse_config(pp_info)
                pp_data = pp(pp_data)
            if torch.is_tensor(pp_data):
                pred = pp_data.cpu().numpy()
            else:
                pred = pp_data
        else:
            pred = y_hat.cpu().numpy()

        if out_fn is None:
            return pred

        # determine output dimension orders
        if out_fn.suffix == ".npy":
            np.save(out_fn, pred)
        else:
            if len(pred.shape) == 2:
                OmeTiffWriter.save(pred, out_fn, dim_order="YX")
            elif len(pred.shape) == 3:
                # 3D output, for 2D data
                if self.spatial_dims == 2:
                    # save as RGB or multi-channel 2D
                    if pred.shape[0] == 3:
                        if out_fn.suffix != ".png":
                            out_fn = out_fn.with_suffix(".png")
                        save_rgb(out_fn, np.moveaxis(pred, 0, -1))
                    else:
                        OmeTiffWriter.save(pred, out_fn, dim_order="CYX")
                elif self.spatial_dims == 3:
                    OmeTiffWriter.save(pred, out_fn, dim_order="ZYX")
                else:
                    raise ValueError("Invalid spatial dimension of pred")
            elif len(pred.shape) == 4:
                if self.spatial_dims == 3:
                    OmeTiffWriter.save(pred, out_fn, dim_order="CZYX")
                elif self.spatial_dims == 2:
                    if pred.shape[0] == 1:
                        if pred.shape[1] == 1:
                            OmeTiffWriter.save(pred[0, 0], out_fn, dim_order="YX")
                        elif pred.shape[1] == 3:
                            if out_fn.suffix != ".png":
                                out_fn = out_fn.with_suffix(".png")
                            save_rgb(
                                out_fn,
                                np.moveaxis(
                                    pred[0,],
                                    0,
                                    -1,
                                ),
                            )
                        else:
                            OmeTiffWriter.save(
                                pred[0,],
                                out_fn,
                                dim_order="CYX",
                            )
                    else:
                        raise ValueError("invalid 4D output for 2d data")
            elif len(pred.shape) == 5:
                assert pred.shape[0] == 1, "error, found non-trivial batch dimension"
                OmeTiffWriter.save(
                    pred[0,],
                    out_fn,
                    dim_order="CZYX",
                )
            else:
                raise ValueError("error in prediction output shape")

    def run_inference(self):

        self.setup_model()
        self.setup_data_processing()
        if "pred_slice2vol" in self.model_cfg.net:

            if self.model_cfg.net["pred_slice2vol"] is not None:
                # handle multiple kind of elements tiff/tif
                if "," in self.data_cfg.inference_input.data_type:
                    types = self.data_cfg.inference_input.data_type.split(",")
                    extensions = [f"*{tipe}" for tipe in types]
                    filenames = sorted(
                        list(
                            itertools.chain.from_iterable(
                                self.data_cfg.inference_input.dir.glob(extension)
                                for extension in extensions
                            )
                        )
                    )
                else:
                    filenames = sorted(
                        self.data_cfg.inference_input.dir.glob(
                            "*" + self.data_cfg.inference_input.data_type
                        )
                    )

                vs_flag = PhysicalPixelSizes(1, 1, 1)
                if "pixel_dim" in self.model_cfg.net["pred_slice2vol"]:
                    if self.model_cfg.net["pred_slice2vol"]["pixel_dim"] is not None:
                        if isinstance(
                            self.model_cfg.net["pred_slice2vol"]["pixel_dim"], str
                        ):
                            vs_flag = "auto"
                        elif isinstance(
                            self.model_cfg.net["pred_slice2vol"]["pixel_dim"], tuple
                        ):
                            if (
                                len(self.model_cfg.net["pred_slice2vol"]["pixel_dim"])
                                == 3
                            ):
                                z, y, x = self.model_cfg.net["pred_slice2vol"][
                                    "pixel_dim"
                                ]
                                vs_flag = PhysicalPixelSizes(z, y, x)
                        elif isinstance(
                            self.model_cfg.net["pred_slice2vol"]["pixel_dim"], list
                        ):
                            if (
                                len(self.model_cfg.net["pred_slice2vol"]["pixel_dim"])
                                == 3
                            ):
                                z, y, x = self.model_cfg.net["pred_slice2vol"][
                                    "pixel_dim"
                                ]
                                vs_flag = PhysicalPixelSizes(z, y, x)

                max_proj = False
                if "max_proj" in self.model_cfg.net["pred_slice2vol"]:
                    if self.model_cfg.net["pred_slice2vol"]["max_proj"] is not None:
                        if isinstance(
                            self.model_cfg.net["pred_slice2vol"]["max_proj"], bool
                        ):
                            max_proj = self.model_cfg.net["pred_slice2vol"]["max_proj"]

                perycites_correction = False
                if "perycites_correction" in self.model_cfg.net["pred_slice2vol"]:
                    if (
                        self.model_cfg.net["pred_slice2vol"]["perycites_correction"]
                        is not None
                    ):
                        perycites_correction = self.model_cfg.net["pred_slice2vol"][
                            "perycites_correction"
                        ]

                print(
                    f"########################################### {len(filenames)} Volume(s) found for prediction ###########################################"
                )

                if (
                    self.model_cfg.net["pred_slice2vol"]["uncertainity_map"]
                    and "prob" not in self.model_cfg.net["func_name"].lower()
                ):
                    print(
                        "##################################################################### Warning #####################################################################"
                    )
                    print(f"Your selected Model is {self.model_cfg.net['func_name']}")
                    print(
                        "If the model is NOT probabilistic the uncertainity map and multiple prediction mode won't have sense"
                    )
                # save post process indicated by the user
                original_postprocess = self.data_cfg.postprocess
                # we set non postprocess we need logits for the model
                self.data_cfg.postprocess = None

                n_trunc = -1
                threshold_um = -1
                border_corr = False
                pert_opt = False
                # handle uncertainity maps generation
                if "uncertainity_map" in self.model_cfg.net["pred_slice2vol"]:
                    if (
                        self.model_cfg.net["pred_slice2vol"]["uncertainity_map"]
                        is not None
                    ):
                        if self.model_cfg.net["pred_slice2vol"]["uncertainity_map"]:

                            if "trunc" in self.model_cfg.net["pred_slice2vol"]:
                                if (
                                    self.model_cfg.net["pred_slice2vol"]["trunc"]
                                    is not None
                                ):
                                    if isinstance(
                                        self.model_cfg.net["pred_slice2vol"]["trunc"],
                                        bool,
                                    ):
                                        if not self.model_cfg.net["pred_slice2vol"][
                                            "trunc"
                                        ]:
                                            n_trunc = -1
                                        else:
                                            n_trunc = 4
                                    else:
                                        if (
                                            type(
                                                self.model_cfg.net["pred_slice2vol"][
                                                    "trunc"
                                                ]
                                            )
                                            is int
                                            and self.model_cfg.net["pred_slice2vol"][
                                                "trunc"
                                            ]
                                            >= 0
                                        ):
                                            n_trunc = self.model_cfg.net[
                                                "pred_slice2vol"
                                            ]["trunc"]
                                        else:
                                            raise ValueError(
                                                f"Unexpected Value for trunc: {self.model_cfg.net['pred_slice2vol']['trunc']}. It should be a positive integer"
                                            )
                                else:
                                    n_trunc = 4
                            else:
                                n_trunc = 4

                            if "threshold" in self.model_cfg.net["pred_slice2vol"]:
                                if (
                                    self.model_cfg.net["pred_slice2vol"]["threshold"]
                                    is not None
                                ):
                                    if isinstance(
                                        self.model_cfg.net["pred_slice2vol"][
                                            "threshold"
                                        ],
                                        bool,
                                    ):
                                        if not self.model_cfg.net["pred_slice2vol"][
                                            "threshold"
                                        ]:
                                            threshold_um = -1
                                        else:
                                            raise ValueError(
                                                f"Unexpected Value for threshold: {self.model_cfg.net['pred_slice2vol']['threshold']}. It should be a positive float"
                                            )
                                    else:
                                        if (
                                            type(
                                                self.model_cfg.net["pred_slice2vol"][
                                                    "threshold"
                                                ]
                                            )
                                            is float
                                            and self.model_cfg.net["pred_slice2vol"][
                                                "threshold"
                                            ]
                                            > 0
                                        ):
                                            threshold_um = self.model_cfg.net[
                                                "pred_slice2vol"
                                            ]["threshold"]
                                        else:
                                            raise ValueError(
                                                f"Unexpected Value for threshold: {self.model_cfg.net['pred_slice2vol']['threshold']}. It should be a positive float"
                                            )
                                else:
                                    threshold_um = -1
                            else:
                                threshold_um = -1

                            if (
                                "border_correction"
                                in self.model_cfg.net["pred_slice2vol"]
                            ):
                                if (
                                    self.model_cfg.net["pred_slice2vol"][
                                        "border_correction"
                                    ]
                                    is not None
                                ):
                                    if isinstance(
                                        self.model_cfg.net["pred_slice2vol"][
                                            "border_correction"
                                        ],
                                        bool,
                                    ):
                                        if not self.model_cfg.net["pred_slice2vol"][
                                            "border_correction"
                                        ]:
                                            border_corr = False
                                        else:
                                            raise ValueError(
                                                f"Unexpected Value for border_correction: {self.model_cfg.net['pred_slice2vol']['border_correction']}."
                                            )
                                    else:
                                        if (
                                            type(
                                                self.model_cfg.net["pred_slice2vol"][
                                                    "border_correction"
                                                ]
                                            )
                                            is int
                                            and self.model_cfg.net["pred_slice2vol"][
                                                "border_correction"
                                            ]
                                            >= 0
                                        ):
                                            if (
                                                self.model_cfg.net["pred_slice2vol"][
                                                    "border_correction"
                                                ]
                                                == 0
                                            ):
                                                border_corr = False
                                            else:
                                                border_corr = [
                                                    self.model_cfg.net[
                                                        "pred_slice2vol"
                                                    ]["border_correction"]
                                                ] * 2
                                        elif (
                                            type(
                                                self.model_cfg.net["pred_slice2vol"][
                                                    "border_correction"
                                                ]
                                            )
                                            is list
                                            and len(
                                                self.model_cfg.net["pred_slice2vol"][
                                                    "border_correction"
                                                ]
                                            )
                                            <= 2
                                        ):
                                            if self.model_cfg.net["pred_slice2vol"][
                                                "border_correction"
                                            ] != [0, 0] and self.model_cfg.net[
                                                "pred_slice2vol"
                                            ][
                                                "border_correction"
                                            ] != [
                                                0
                                            ]:
                                                if (
                                                    len(
                                                        self.model_cfg.net[
                                                            "pred_slice2vol"
                                                        ]["border_correction"]
                                                    )
                                                    == 1
                                                ):
                                                    border_corr = (
                                                        self.model_cfg.net[
                                                            "pred_slice2vol"
                                                        ]["border_correction"]
                                                        * 2
                                                    )
                                                else:
                                                    border_corr = self.model_cfg.net[
                                                        "pred_slice2vol"
                                                    ]["border_correction"]
                                            else:
                                                border_corr = False
                                        else:
                                            raise ValueError(
                                                f"Unexpected Value for border_correction: {self.model_cfg.net['pred_slice2vol']['border_correction']}."
                                            )
                                else:
                                    border_corr = False
                            else:
                                border_corr = False

                            if self.model_cfg.net["pred_slice2vol"]["n_samples"] <= 1:
                                print(
                                    "Number of samples are less or equal to 1 more are required for uncertainty generation"
                                )
                                print(
                                    "Automatically 5 samples will use to uncetainity calculation"
                                )
                                self.model_cfg.net["pred_slice2vol"]["n_samples"] = 5

                            if "var_reductor" in self.model_cfg.net["pred_slice2vol"]:
                                if (
                                    self.model_cfg.net["pred_slice2vol"]["var_reductor"]
                                    is None
                                ):
                                    self.model_cfg.net["pred_slice2vol"][
                                        "var_reductor"
                                    ] = True
                            else:
                                self.model_cfg.net["pred_slice2vol"][
                                    "var_reductor"
                                ] = True

                            if "relative_MI" in self.model_cfg.net["pred_slice2vol"]:
                                if (
                                    self.model_cfg.net["pred_slice2vol"]["relative_MI"]
                                    is None
                                ):
                                    self.model_cfg.net["pred_slice2vol"][
                                        "relative_MI"
                                    ] = True
                            else:
                                self.model_cfg.net["pred_slice2vol"][
                                    "relative_MI"
                                ] = True

                            if "compute_mode" in self.model_cfg.net["pred_slice2vol"]:
                                if (
                                    self.model_cfg.net["pred_slice2vol"]["compute_mode"]
                                    is None
                                ):
                                    self.model_cfg.net["pred_slice2vol"][
                                        "compute_mode"
                                    ] = "mutual_inf"
                            else:
                                self.model_cfg.net["pred_slice2vol"][
                                    "compute_mode"
                                ] = "mutual_inf"

                            if "estabilizer" in self.model_cfg.net["pred_slice2vol"]:
                                if (
                                    self.model_cfg.net["pred_slice2vol"]["estabilizer"]
                                    is None
                                ):
                                    self.model_cfg.net["pred_slice2vol"][
                                        "estabilizer"
                                    ] = False
                            else:
                                self.model_cfg.net["pred_slice2vol"][
                                    "estabilizer"
                                ] = False

                            if "pertubations" in self.model_cfg.net["pred_slice2vol"]:
                                if (
                                    self.model_cfg.net["pred_slice2vol"]["pertubations"]
                                    is not None
                                ):
                                    if isinstance(
                                        self.model_cfg.net["pred_slice2vol"][
                                            "pertubations"
                                        ],
                                        str,
                                    ):
                                        pert_opt = True
                                    elif isinstance(
                                        self.model_cfg.net["pred_slice2vol"][
                                            "pertubations"
                                        ],
                                        bool,
                                    ):
                                        pert_opt = True
                                    elif isinstance(
                                        self.model_cfg.net["pred_slice2vol"][
                                            "pertubations"
                                        ],
                                        list,
                                    ):
                                        if (
                                            len(
                                                self.model_cfg.net["pred_slice2vol"][
                                                    "pertubations"
                                                ]
                                            )
                                            != 0
                                        ):
                                            pert_opt = True
                    else:
                        self.model_cfg.net["pred_slice2vol"]["uncertainity_map"] = False
                else:
                    self.model_cfg.net["pred_slice2vol"]["uncertainity_map"] = False

                # handle multi pred generation
                if "multi_pred_mode" in self.model_cfg.net["pred_slice2vol"]:
                    if (
                        self.model_cfg.net["pred_slice2vol"]["multi_pred_mode"]
                        is not None
                    ):
                        if (
                            self.model_cfg.net["pred_slice2vol"][
                                "multi_pred_mode"
                            ].lower()
                            != "single"
                        ):
                            if self.model_cfg.net["pred_slice2vol"]["n_samples"] <= 1:
                                print(
                                    "Number of samples are less or equal to 1 more are required for multi prediction usage"
                                )
                                print("Automatically 5 samples will use to prediction")
                                self.model_cfg.net["pred_slice2vol"]["n_samples"] = 5
                    else:
                        self.model_cfg.net["pred_slice2vol"][
                            "multi_pred_mode"
                        ] = "single"

                else:
                    self.model_cfg.net["pred_slice2vol"]["multi_pred_mode"] = "single"

                if "jupyter" in self.model_cfg.net["pred_slice2vol"]:
                    from tqdm.notebook import tqdm
                else:
                    from tqdm import tqdm

                for fn in tqdm(filenames, desc="Prediction for the Volume"):
                    try:
                        img = BioImage(fn, reader=bioio_tifffile.Reader).get_image_data(
                            self.data_cfg.inference_input.reader_params[
                                "dimension_order_out"
                            ],
                            T=self.data_cfg.inference_input.reader_params["T"],
                        )
                    except Exception:
                        try:
                            img = BioImage(fn).get_image_data(
                                self.data_cfg.inference_input.reader_params[
                                    "dimension_order_out"
                                ],
                                T=self.data_cfg.inference_input.reader_params["T"],
                            )
                        except Exception as e:
                            print(f"Error: {e}")
                            print(
                                f"Image {fn} failed at read process check the format."
                            )

                    voxel_sizes = PhysicalPixelSizes(1, 1, 1)
                    if vs_flag == "auto":
                        pps = getattr(BioImage(fn), "physical_pixel_sizes", None)
                        if pps is None:
                            voxel_sizes = PhysicalPixelSizes(None, None, None)
                        elif isinstance(pps, tuple):
                            voxel_sizes = pps  # tuple like (Z, Y, X)
                        else:
                            voxel_sizes = (
                                getattr(pps, "Z", None),
                                getattr(pps, "Y", None),
                                getattr(pps, "X", None),
                            )
                            voxel_sizes = PhysicalPixelSizes(
                                voxel_sizes[0], voxel_sizes[1], voxel_sizes[2]
                            )

                        voxel_sizes = [
                            1.0 if v is None else float(v) for v in voxel_sizes
                        ]
                        voxel_sizes = PhysicalPixelSizes(
                            voxel_sizes[0], voxel_sizes[1], voxel_sizes[2]
                        )
                    else:
                        voxel_sizes = vs_flag

                    if max_proj:
                        img = np.max(img, axis=1)
                        img = np.expand_dims(img, axis=1)

                    out_list = []
                    uncertainity_map = []
                    n = fn.name
                    if len(img.shape) == 3:
                        # chanel dummy
                        img = img[None, ...]

                    for zz in tqdm(
                        range(img.shape[1]), desc="infering slice", leave=False
                    ):
                        samplesz = []
                        im_input = img[:, zz, :, :]
                        for i in range(
                            self.model_cfg.net["pred_slice2vol"]["n_samples"]
                        ):

                            if pert_opt and i != 0:
                                inp = perturb_image(
                                    im_input,
                                    self.model_cfg.net["pred_slice2vol"][
                                        "pertubations"
                                    ],
                                )
                            else:
                                inp = im_input

                            logits = self.process_one_image(inp)
                            samplesz.append(np.squeeze(logits))

                        if (
                            self.model_cfg.net["pred_slice2vol"][
                                "multi_pred_mode"
                            ].lower()
                            == "single"
                        ):
                            seg = samplesz[0]
                            seg = seg[None, ...]
                        elif (
                            self.model_cfg.net["pred_slice2vol"][
                                "multi_pred_mode"
                            ].lower()
                            == "max"
                        ):
                            seg = max_prediction(samplesz)
                            seg = seg[None, ...]
                        elif (
                            self.model_cfg.net["pred_slice2vol"][
                                "multi_pred_mode"
                            ].lower()
                            == "mean"
                        ):
                            seg = mean_prediction(samplesz)
                            seg = seg[None, ...]
                        elif (
                            self.model_cfg.net["pred_slice2vol"][
                                "multi_pred_mode"
                            ].lower()
                            == "variance"
                        ):
                            seg = variance_prediction(samplesz)
                            seg = seg[None, ...]
                        elif (
                            self.model_cfg.net["pred_slice2vol"][
                                "multi_pred_mode"
                            ].lower()
                            == "sum"
                        ):
                            seg = add_prediction(samplesz)
                            seg = seg[None, ...]
                        else:
                            raise ValueError(
                                f"{self.model_cfg.net['pred_slice2vol']['multi_pred_mode']} is not a valid method for multi prediction use."
                            )

                        if original_postprocess is not None:
                            pp_data = seg
                            for pp_info in original_postprocess:
                                pp = parse_config(pp_info)
                                pp_data = pp(pp_data)
                            if torch.is_tensor(pp_data):
                                seg = pp_data.cpu().numpy()
                            else:
                                seg = pp_data
                        out_list.append(np.squeeze(seg))
                        if self.model_cfg.net["pred_slice2vol"]["uncertainity_map"]:
                            Uc_zmap = Extract_Uncertainty_Maps(
                                logits_samples=samplesz,
                                compute_mode=self.model_cfg.net["pred_slice2vol"][
                                    "compute_mode"
                                ],
                                relative_MI=self.model_cfg.net["pred_slice2vol"][
                                    "relative_MI"
                                ],
                                var_reductor=self.model_cfg.net["pred_slice2vol"][
                                    "var_reductor"
                                ],
                                estabilizer=self.model_cfg.net["pred_slice2vol"][
                                    "estabilizer"
                                ],
                            )
                            uncertainity_map.append(Uc_zmap)

                            # elimina _IM del name para vessqc
                            if "_IM" in n:
                                n = n.replace("_IM", "")

                    seg_full = np.stack(out_list, axis=0)
                    if self.model_cfg.net["pred_slice2vol"]["uncertainity_map"]:
                        UM_full = np.stack(uncertainity_map, axis=0)

                    if "remove_object_size" in self.model_cfg.net["pred_slice2vol"]:
                        if (
                            self.model_cfg.net["pred_slice2vol"]["remove_object_size"]
                            is not None
                        ):
                            seg_full = Remove_objects(
                                seg_full=seg_full,
                                n_classes=self.model_cfg.net["pred_slice2vol"][
                                    "n_class_correction"
                                ],
                                remove_object_size=self.model_cfg.net["pred_slice2vol"][
                                    "remove_object_size"
                                ],
                                voxel_sizes=tuple(voxel_sizes),
                            )

                    if "hole_size_threshold" in self.model_cfg.net["pred_slice2vol"]:
                        if (
                            self.model_cfg.net["pred_slice2vol"]["hole_size_threshold"]
                            is not None
                        ):
                            seg_full = Hole_Correction(
                                seg_full=seg_full,
                                n_classes=self.model_cfg.net["pred_slice2vol"][
                                    "n_class_correction"
                                ],
                                hole_size_threshold=self.model_cfg.net[
                                    "pred_slice2vol"
                                ]["hole_size_threshold"],
                                voxel_sizes=tuple(voxel_sizes),
                            )

                    if "min_thickness_list" in self.model_cfg.net["pred_slice2vol"]:
                        if (
                            self.model_cfg.net["pred_slice2vol"]["min_thickness_list"]
                            is not None
                        ):
                            seg_full = Thickness_Corretion(
                                seg_full=seg_full,
                                n_classes=self.model_cfg.net["pred_slice2vol"][
                                    "n_class_correction"
                                ],
                                min_thickness_physical=self.model_cfg.net[
                                    "pred_slice2vol"
                                ]["min_thickness_list"],
                                voxel_sizes=tuple(voxel_sizes),
                            )

                    if perycites_correction:
                        seg_full = Perycites_correction(seg_full=seg_full)

                    if ".tiff" in n:

                        out_fn = self.data_cfg.inference_output.path / n.replace(
                            ".tiff", "_segPred.tiff"
                        )
                        UM_out_fn = self.data_cfg.inference_output.path / n.replace(
                            ".tiff", "_uncertainty.tiff"
                        )
                    elif ".tif" in n:
                        out_fn = self.data_cfg.inference_output.path / n.replace(
                            ".tif", "_segPred.tif"
                        )
                        UM_out_fn = self.data_cfg.inference_output.path / n.replace(
                            ".tif", "_uncertainty.tif"
                        )

                    OmeTiffWriter.save(
                        data=seg_full,
                        uri=out_fn,
                        dim_order="ZYX",
                        physical_pixel_sizes=voxel_sizes,
                        physical_pixel_units="micron",
                    )

                    if self.model_cfg.net["pred_slice2vol"]["uncertainity_map"]:

                        if n_trunc >= 0:
                            UM_full = np.trunc(UM_full * (10**n_trunc)) / (10**n_trunc)

                        if threshold_um >= 0:
                            UM_full[UM_full < threshold_um] = 0

                        if border_corr:
                            nx, ny = border_corr
                            nx = abs(nx)
                            ny = abs(ny)
                            Z, X, Y = UM_full.shape
                            UM_full[:, : nx + 1, :] = 0
                            UM_full[:, X - (nx + 1) :, :] = 0
                            UM_full[:, :, : (ny + 1)] = 0
                            UM_full[:, :, Y - (ny + 1) :] = 0

                        if self.model_cfg.net["pred_slice2vol"]["var_reductor"]:
                            OmeTiffWriter.save(UM_full, UM_out_fn, dim_order="ZYX")

                        else:
                            UM_full_CZYX = np.moveaxis(UM_full, 1, 0)
                            OmeTiffWriter.save(
                                UM_full_CZYX, UM_out_fn, dim_order="CZYX"
                            )

            else:
                raise ValueError("Please provide params for the volumetric prediction.")



import torch.nn.functional as F
import torch
import torch.nn as nn
import numpy as np
from scipy.stats import linregress
import math


class Slice_windows(nn.Module):
    """
    A PyTorch module to count windows with ones or calculate the average Shannon entropy
    of windows in an input tensor.

    Args:
        num_kernels (int): The number of kernel sizes to use. Must be between 1 and 10.
                           The kernel sizes will be 2^1, 2^2, ..., 2^num_kernels.
        mode (str): The operational mode of the module.
                    - "classic": Counts the number of windows that contain at least one '1'.
                    - "entropy": Calculates the average Shannon entropy for each window.
                                 H = -p0 * log2(p0) - p1 * log2(p1), where p0 and p1 are
                                 the probabilities of 0s and 1s in the window, respectively.
        to_binary (bool): If True, the input tensor will be converted to a binary tensor
                          where all values > 0 become 1, and 0 remains 0. Defaults to False.
    """

    def __init__(
        self, num_kernels: int, mode: str = "classic", to_binary: bool = False
    ):
        super().__init__()
        if not (1 <= num_kernels <= 10):
            raise ValueError(
                "num_kernels must be between 1 and 10 to avoid excessively large kernels."
            )
        self.num_kernels = num_kernels
        # Kernel sizes are powers of 2, from 2^1 to 2^num_kernels
        self.kernel_sizes = [2**i for i in range(1, num_kernels + 1)]

        if mode not in ["classic", "entropy"]:
            raise ValueError(f"Mode must be 'classic' or 'entropy', but got '{mode}'.")
        self.mode = mode
        self.to_binary = to_binary

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Processes the input tensor based on the selected mode.

        Args:
            x (torch.Tensor): The input tensor, expected to be 4D (Batch, Channels, H, W).

        Returns:
            torch.Tensor: A 2D tensor (Batch, num_kernels) containing the results.
                          If mode is "classic", it contains the count of windows with ones.
                          If mode is "entropy", it contains the average entropy of windows.
        """
        if x.dim() != 4:
            raise ValueError(
                f"Input must be a 4D tensor (Batch, Channels, H, W), but got {x.dim()}D."
            )

        # Convert to binary if to_binary is True
        if self.to_binary:
            x = (
                x > 0
            ).float()  # Convert to float to maintain tensor type for subsequent operations
            # Values > 0 become 1.0, 0 remains 0.0

        batch_size, channels, H, W = x.shape
        # Initialize results tensor to store output for each batch and kernel size
        results = torch.zeros(batch_size, self.num_kernels, device=x.device)

        # Iterate through each defined kernel size
        for i, kernel_size in enumerate(self.kernel_sizes):
            # If kernel size is larger than image dimensions, skip and set result to 0
            if kernel_size > H or kernel_size > W:
                results[:, i] = (
                    0.0  # Use 0.0 for consistency with entropy which can be float
                )
                continue

            # Calculate the total number of elements in a single window
            elements_in_window = kernel_size * kernel_size * channels

            # Process each item in the batch
            for b in range(batch_size):
                # Unfold the input tensor into overlapping windows.
                # x[b:b+1] is used to maintain the batch dimension for F.unfold
                unfolded_windows = F.unfold(
                    x[b : b + 1],
                    kernel_size=(kernel_size, kernel_size),
                    stride=(kernel_size, kernel_size),
                )

                # Transpose and squeeze to get shape (num_windows, elements_in_window)
                # where num_windows is the total number of non-overlapping windows
                unfolded_windows = unfolded_windows.transpose(1, 2).squeeze(0)

                if self.mode == "classic":
                    # Classic mode: Count windows containing at least one '1'
                    # Sum elements in each window, if sum > 0, it has at least one '1'
                    windows_with_ones = (unfolded_windows.sum(dim=1) > 0).float()
                    current_count = torch.sum(windows_with_ones).item()
                    results[b, i] = current_count

                elif self.mode == "entropy":
                    # Entropy mode: Calculate Shannon entropy for each window
                    entropies = []
                    # Iterate over each window
                    for window in unfolded_windows:
                        # Count number of ones in the current window
                        num_ones = window.sum().item()
                        # Count number of zeros in the current window
                        num_zeros = elements_in_window - num_ones

                        # Calculate probabilities
                        p1 = num_ones / elements_in_window
                        p0 = num_zeros / elements_in_window

                        # Shannon entropy calculation
                        # H = -p0 * log2(p0) - p1 * log2(p1)
                        # Handle log2(0) case: if p is 0, p * log2(p) is considered 0
                        entropy = 0.0
                        if p0 > 0:
                            entropy -= p0 * math.log2(p0)
                        if p1 > 0:
                            entropy -= p1 * math.log2(p1)
                        entropies.append(entropy)

                    # Calculate the average entropy for the current batch item and kernel size
                    if entropies:  # Ensure there are entropies to average
                        results[b, i] = sum(entropies) / len(entropies)
                    else:
                        results[b, i] = 0.0  # No windows, so entropy is 0

        return results


class FractalDimension(nn.Module):
    def __init__(
        self, num_kernels: int, mode: str = "classic", to_binary: bool = False
    ):
        """
        Initializes the layer to estimate a scaling exponent based on box characteristics.

        Args:
            num_kernels (int): The number of kernels to use for the underlying box-counting/entropy layer.
            mode (str): The operational mode for the underlying Slice_windows layer.
                        - "classic": The estimator calculates the traditional box-counting fractal dimension.
                        - "entropy": The estimator calculates a scaling exponent related to how
                                     average Shannon entropy changes with box size.
            to_binary (bool): If True, the input tensor will be converted to a binary tensor
                              where all values > 0 become 1, and 0 remains 0. This parameter
                              is passed directly to the Slice_windows layer. Defaults to False.
        """
        super().__init__()
        # Pass the mode and to_binary directly to the Slice_windows layer
        self.count_layer = Slice_windows(
            num_kernels=num_kernels, mode=mode, to_binary=to_binary
        )
        self.kernel_sizes = (
            self.count_layer.kernel_sizes
        )  # Get kernel sizes from the sub-layer

        # Store the mode to clarify the output interpretation
        self.mode = mode
        self.to_binary = to_binary  # Store it for potential external inspection, though not strictly used internally in forward

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculates a scaling exponent for each image in the batch.

        Args:
            x (torch.Tensor): The input image. If `to_binary` was set to True during initialization,
                              this input will be converted to binary internally by `Slice_windows`.
                              Expected format: (Batch_size, Channels, Height, Width)

        Returns:
            torch.Tensor: A tensor of size (Batch_size,) where each element
                          is the estimated scaling exponent for the corresponding image.
                          If mode is "classic", this is the box-counting fractal dimension.
                          If mode is "entropy", this is an entropy-based scaling exponent.
                          Returns 0.0 if regression cannot be performed (e.g., not enough valid points).
        """
        # Get the counts/entropies from the underlying layer based on the selected mode
        # The to_binary conversion happens inside self.count_layer if self.to_binary was True
        results_per_kernel = self.count_layer(x)

        batch_size = x.shape[0]
        scaling_exponents = torch.zeros(
            batch_size, device=x.device, dtype=torch.float32
        )

        # Calculate the scaling exponent for each image in the batch
        for b in range(batch_size):
            current_results = results_per_kernel[b].cpu().numpy()

            # Calculate inverse kernel sizes for the x-axis of the log-log plot
            inverse_kernel_sizes = np.array([1 / k for k in self.kernel_sizes])

            # Filter valid points for regression: only consider points where the result is greater than 0
            # This avoids issues with log(0) which would lead to -infinity
            valid_indices = current_results > 0

            if (
                np.sum(valid_indices) < 2
            ):  # Need at least 2 points for a linear regression
                scaling_exponents[b] = 0.0  # Cannot estimate scaling exponent
                continue

            # Take the logarithm of the results and inverse kernel sizes
            if self.mode == "classic":
                log_results = np.log(current_results[valid_indices])
                log_inverse_kernel_sizes = np.log(inverse_kernel_sizes[valid_indices])
            if self.mode == "entropy":
                log_results = current_results[valid_indices]
                log_inverse_kernel_sizes = np.log(inverse_kernel_sizes[valid_indices])

            try:
                # Perform linear regression: the slope is the scaling exponent
                slope, intercept, r_value, p_value, std_err = linregress(
                    log_inverse_kernel_sizes, log_results
                )
                scaling_exponents[b] = torch.tensor(
                    slope, dtype=torch.float32, device=x.device
                )
            except ValueError:
                # This can happen if there are issues with the input to linregress (e.g., all same values)
                scaling_exponents[b] = 0.0
            except Exception as e:
                print(f"Error during linregress for batch {b}: {e}")
                scaling_exponents[b] = 0.0

        return scaling_exponents



import numpy as np
from typing import Dict
from pathlib import Path
from bioio.writers import OmeTiffWriter
import lightning as pl
from mmv_im2im.postprocessing.embedseg_cluster import generate_instance_clusters
from mmv_im2im.utils.embedseg_utils import prepare_embedseg_tensor
from mmv_im2im.utils.model_utils import init_weights
import torch
from monai.inferers import sliding_window_inference

from mmv_im2im.utils.misc import (
    parse_config,
    parse_config_func,
    parse_config_func_without_params,
)
from mmv_im2im.utils.metrics import simplified_instance_IoU


class Model(pl.LightningModule):
    def __init__(self, model_info_xx: Dict, train: bool = True, verbose: bool = False):
        super().__init__()
        self.verbose = verbose

        # for training, we need the clustering parameters for the validation step,
        # the same set of parameters as in the criterion, except "foreground_weight", so
        # pop it out
        if train:
            self.clustering_params = model_info_xx.criterion["params"].copy()
            self.clustering_params.pop("foreground_weight")
            if "use_costmap" in self.clustering_params:
                self.clustering_params.pop("use_costmap")

        self.net = parse_config(model_info_xx.net)

        if model_info_xx.net["func_name"].startswith("BranchedERFNet"):
            if train:
                # if not train, not need to init
                self.net.init_output(model_info_xx.criterion["params"]["n_sigma"])
        else:
            init_weights(self.net, init_type="kaiming")

        self.model_info = model_info_xx
        if train:
            extra_clustering_param = [
                "min_mask_sum",
                "min_unclustered_sum",
                "min_object_size",
            ]
            for cp in extra_clustering_param:
                if cp in model_info_xx.criterion["params"]:
                    model_info_xx.criterion["params"].pop(cp)

            self.criterion = parse_config(model_info_xx.criterion)
            self.optimizer_func = parse_config_func(model_info_xx.optimizer)

    def configure_optimizers(self):
        optimizer = self.optimizer_func(self.parameters())
        if self.model_info.scheduler is None:
            return optimizer
        else:
            scheduler_func = parse_config_func_without_params(self.model_info.scheduler)
            lr_scheduler = scheduler_func(
                optimizer, **self.model_info.scheduler["params"]
            )
            return {"optimizer": optimizer, "lr_scheduler": lr_scheduler}

    def forward(self, x):
        return self.net(x)

    def run_step(self, batch, validation_stage, save_path: str = None):
        im = batch["IM"]
        instances = batch["GT"]

        if len(im.size()) == 4:  # "BCYX"
            spatial_dim = 2
        elif len(im.size()) == 5:  # "BCZYX"
            spatial_dim = 3

        use_costmap = False
        if "CM" in batch:
            use_costmap = True
            costmap = batch["CM"]

        # forward
        if validation_stage:
            # generate predictions
            if (
                self.model_info.model_extra is not None
                and "validation_sliding_windows" in self.model_info.model_extra
            ):
                output = sliding_window_inference(
                    inputs=im,
                    predictor=self.net,
                    device=torch.device("cpu"),
                    **self.model_info.model_extra["validation_sliding_windows"],
                )
                # move back to CUDA
                output = output.cuda()
            else:
                output = self.net(im)

            # generate instance segmentation
            instances_map = generate_instance_clusters(output, **self.clustering_params)

            # calculate simplified IoU loss for validation
            instances = instances.detach().cpu().numpy()
            if len(instances.shape) > spatial_dim:
                instances = np.squeeze(instances)

            if use_costmap:
                sIoU = simplified_instance_IoU(
                    instances, instances_map, costmap.detach().cpu().numpy() == 0
                )
            else:
                sIoU = simplified_instance_IoU(instances, instances_map)

            loss = sIoU
        else:
            output = self(im)

            # decide if CL and CE are already generated
            if "CL" in batch and "CE" in batch:
                class_labels = batch["CL"]
                center_images = batch["CE"]
            else:
                class_labels, center_images = prepare_embedseg_tensor(
                    instances, spatial_dim, self.model_info.model_extra["center_method"]
                )

            if use_costmap:
                loss = self.criterion(
                    output, instances, class_labels, center_images, costmap
                )
            else:
                loss = self.criterion(output, instances, class_labels, center_images)
            loss = loss.mean()

        if save_path is not None:
            current_stage = "val"
            # get instance segmentation if not yet
            if not validation_stage:
                current_stage = "train"
                instances_map = generate_instance_clusters(
                    output, **self.clustering_params
                )
                # remove the batch dimension
                gt = instances.detach().cpu().numpy()[0,]
            else:
                # add back the C dimension
                gt = np.expand_dims(instances, axis=0)
            if len(im.size()) == 4:
                dim_order = "CYX"
            elif len(im.size()) == 5:
                dim_order = "CZYX"

            # save raw image
            out_fn = save_path / f"{self.current_epoch}_{current_stage}_raw.tiff"
            OmeTiffWriter.save(
                im.detach().cpu().numpy()[0,],
                out_fn,
                dim_order=dim_order,
            )

            # save ground truth
            out_fn = save_path / f"{self.current_epoch}_{current_stage}_gt.tiff"
            OmeTiffWriter.save(gt, out_fn, dim_order=dim_order)

            # # save instance segmentation
            out_fn = save_path / f"{self.current_epoch}_{current_stage}_seg.tiff"
            OmeTiffWriter.save(instances_map, out_fn, dim_order=dim_order[1:])
        return loss

    def training_step(self, batch, batch_idx):
        if self.verbose and batch_idx == 0:
            # check if the log path exists, if not create one
            log_dir = Path(self.trainer.log_dir)
            log_dir.mkdir(parents=True, exist_ok=True)

            loss = self.run_step(batch, validation_stage=False, save_path=log_dir)
        else:
            loss = self.run_step(batch, validation_stage=False)
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        return loss

    def validation_step(self, batch, batch_idx):
        if self.verbose and batch_idx == 0:
            # check if the log path exists, if not create one
            log_dir = Path(self.trainer.log_dir)
            log_dir.mkdir(parents=True, exist_ok=True)
            loss = self.run_step(batch, validation_stage=False, save_path=log_dir)
        else:
            loss = self.run_step(batch, validation_stage=False)
        self.log(
            "val_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        return loss



########################################################################
# the code was adapted from https://github.com/wolny/pytorch-3dunet
# with MIT license
########################################################################
from typing import List
import torch
from torch import nn


def expand_as_one_hot(input, C, ignore_index=None):
    """
    Converts NxSPATIAL label image to NxCxSPATIAL, where each label gets converted to
    its corresponding one-hot vector. It is assumed that the batch dimension is present.
    Args:
        input (torch.Tensor): 3D/4D input image
        C (int): number of channels/labels
        ignore_index (int): ignore index to be kept during the expansion
    Returns:
        4D/5D output torch.Tensor (NxCxSPATIAL)
    """
    assert (
        input.dim() == 4 or input.dim() == 3
    ), "only support 3D/4D (i.e., 2D/3D images)"

    # expand the input tensor to Nx1xSPATIAL before scattering
    input = input.unsqueeze(1)
    # create output tensor shape (NxCxSPATIAL)
    shape = list(input.size())
    shape[1] = C

    if ignore_index is not None:
        # create ignore_index mask for the result
        mask = input.expand(shape) == ignore_index
        # clone the src tensor and zero out ignore_index in the input
        input = input.clone()
        input[input == ignore_index] = 0
        # scatter to get the one-hot tensor
        result = torch.zeros(shape).to(input.device).scatter_(1, input, 1)
        # bring back the ignore_index in the result
        result[mask] = ignore_index
        return result
    else:
        # scatter to get the one-hot tensor
        return torch.zeros(shape).to(input.device).scatter_(1, input, 1)


class PixelWiseCrossEntropyLoss(nn.Module):
    def __init__(self, class_weights: List = None, ignore_index=None, one_hot_gt=False):
        super(PixelWiseCrossEntropyLoss, self).__init__()
        if class_weights is None:
            self.class_weights = None
        else:
            self.class_weights = torch.FloatTensor(class_weights)
        self.ignore_index = ignore_index
        self.log_softmax = nn.LogSoftmax(dim=1)
        self.one_hot = one_hot_gt

    def forward(self, input, target, weights=None):
        # input: N x C x Spatial
        # target: N x 1 x Spatial (not one hot) or N x C x Spatial (one_hot)
        # weights: N x 1 x Spatial

        if weights is None:
            weights = torch.ones_like(target)

        # normalize the input
        log_probabilities = self.log_softmax(input)

        # convert to one hot
        if not self.one_hot:
            # need to remove the dummy C dimension before converting to one hot
            target = target.squeeze(1)
            target = expand_as_one_hot(
                target, C=input.size()[1], ignore_index=self.ignore_index
            )
        # expand weights
        weights = weights.expand_as(input)

        # create default class_weights if None
        if self.class_weights is None:
            class_weights = torch.ones(input.size()[1]).float().to(input.device)
        else:
            class_weights = self.class_weights.to(input.device)

        # resize class_weights to be broadcastable into the weights
        class_weights = class_weights.view(1, -1, 1, 1, 1)

        # multiply weights tensor by class weights
        weights = class_weights * weights

        # compute the losses
        result = -weights * target * log_probabilities
        # average the losses
        return result.mean()



from torch.nn import init


def init_weights(net, init_type="kaiming", init_gain=0.02):
    """Initialize network weights.
    Parameters:
        net (network)   -- network to be initialized
        init_type (str) -- the name of an initialization method:
                           normal | xavier | kaiming | orthogonal
        init_gain (float)    -- scaling factor for normal, xavier and orthogonal.
    We use 'normal' in the original pix2pix and CycleGAN paper. But xavier and kaiming
    might work better for some applications. Feel free to try yourself.
    """

    def init_func(m):  # define the initialization function
        classname = m.__class__.__name__
        if hasattr(m, "weight"):
            if classname.find("Conv") != -1 or classname.find("Linear") != -1:
                if init_type == "normal":
                    init.normal_(m.weight.data, 0.0, init_gain)
                elif init_type == "xavier":
                    init.xavier_normal_(m.weight.data, gain=init_gain)
                elif init_type == "kaiming":
                    init.kaiming_normal_(m.weight.data, a=0, mode="fan_in")
                elif init_type == "orthogonal":
                    init.orthogonal_(m.weight.data, gain=init_gain)
                else:
                    raise NotImplementedError(
                        "initialization method [%s] is not implemented" % init_type
                    )
                if hasattr(m, "bias") and m.bias is not None:
                    init.constant_(m.bias.data, 0.0)
            elif classname.find("BatchNorm") != -1:
                # BatchNorm Layer's weight is not a matrix; only normal distribution.
                init.normal_(m.weight.data, 1.0, init_gain)
                init.constant_(m.bias.data, 0.0)

    print("initialize network with %s" % init_type)
    net.apply(init_func)  # apply the initialization function <init_func>


def state_dict_simplification(state_dict, cut: str = "net."):
    """remove the prefix in all keys of state_dict"""
    state_dict_copy = state_dict.copy()
    for old_key in state_dict.keys():
        new_key = old_key.removeprefix(cut)
        state_dict_copy[new_key] = state_dict_copy.pop(old_key)
    return state_dict_copy



import torch
import torch.nn as nn


class Bottleneck(nn.Module):
    def __init__(self, nc):
        super().__init__()
        self.main_block = nn.Sequential(
            nn.BatchNorm2d(num_features=4 * nc),
            nn.ReLU(),
            nn.Conv2d(in_channels=4 * nc, out_channels=nc, kernel_size=1, stride=1),
            nn.BatchNorm2d(num_features=nc),
            nn.ReLU(),
            nn.Conv2d(
                in_channels=nc, out_channels=nc, kernel_size=3, stride=1, padding=1
            ),
            nn.BatchNorm2d(num_features=nc),
            nn.ReLU(),
            nn.Conv2d(in_channels=nc, out_channels=4 * nc, kernel_size=1, stride=1),
        )

    def forward(self, x):
        return x + self.main_block(x)


class Bottleneck_transition(nn.Module):
    def __init__(self, in_channels, nc):
        super().__init__()
        self.main_block = nn.Sequential(
            nn.BatchNorm2d(num_features=in_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                in_channels=in_channels, out_channels=nc, kernel_size=1, stride=1
            ),
            nn.BatchNorm2d(num_features=nc),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                in_channels=nc, out_channels=nc, kernel_size=3, stride=1, padding=1
            ),
            nn.BatchNorm2d(num_features=nc),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=nc, out_channels=4 * nc, kernel_size=1, stride=1),
        )
        self.projection = nn.Sequential(
            nn.BatchNorm2d(num_features=in_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                in_channels=in_channels, out_channels=4 * nc, kernel_size=1, stride=1
            ),
        )

    def forward(self, x):
        return self.projection(x) + self.main_block(x)


class SuggestiveAnnotationModel(torch.nn.Module):
    def __init__(self, in_channels=1, out_channels=2, num_feature=32):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.p1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.block1 = nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=num_feature,
                kernel_size=3,
                padding=1,
            ),
            nn.BatchNorm2d(num_features=num_feature),
            nn.ReLU(),
            nn.Conv2d(
                in_channels=num_feature,
                out_channels=num_feature,
                kernel_size=3,
                padding=1,
            ),
            nn.BatchNorm2d(num_features=num_feature),
            nn.ReLU(),
        )

        # part 2 (nc = 64)
        self.p2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.block2 = nn.Sequential(
            Bottleneck_transition(in_channels=num_feature * 1, nc=num_feature * 2),
            Bottleneck(nc=num_feature * 2),
        )

        # part 3 (nc = 128)
        self.p3 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.block3 = nn.Sequential(
            Bottleneck_transition(in_channels=num_feature * 8, nc=num_feature * 4),
            Bottleneck(nc=num_feature * 4),
        )

        # part 4 (nc = 256)
        self.p4 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.block4 = nn.Sequential(
            Bottleneck_transition(in_channels=num_feature * 16, nc=num_feature * 8),
            Bottleneck(nc=num_feature * 8),
        )

        # part 5 (nc = 256)
        self.p5 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.block5 = nn.Sequential(
            Bottleneck_transition(in_channels=num_feature * 32, nc=num_feature * 8),
            Bottleneck(nc=num_feature * 8),
        )

        # part 6 (nc = 256)
        self.p6 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.block6 = nn.Sequential(
            Bottleneck_transition(in_channels=num_feature * 32, nc=num_feature * 8),
            Bottleneck(nc=num_feature * 8),
        )

        # decoder for part 6
        self.dc6 = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=num_feature * 32,
                out_channels=num_feature * 16,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 16,
                out_channels=num_feature * 8,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 8,
                out_channels=num_feature * 4,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 4,
                out_channels=num_feature * 2,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 2,
                out_channels=num_feature * 1,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 1,
                out_channels=out_channels,
                kernel_size=2,
                stride=2,
            ),
        )

        # decoder for part 5
        self.dc5 = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=num_feature * 32,
                out_channels=num_feature * 16,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 16,
                out_channels=num_feature * 8,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 8,
                out_channels=num_feature * 4,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 4,
                out_channels=num_feature * 2,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 2,
                out_channels=out_channels,
                kernel_size=2,
                stride=2,
            ),
        )

        # decoder for part 4
        self.dc4 = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=num_feature * 32,
                out_channels=num_feature * 16,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 16,
                out_channels=num_feature * 8,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 8,
                out_channels=num_feature * 4,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 4,
                out_channels=out_channels,
                kernel_size=2,
                stride=2,
            ),
        )

        # decoder for part 3
        self.dc3 = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=num_feature * 16,
                out_channels=num_feature * 8,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 8,
                out_channels=num_feature * 4,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 4,
                out_channels=out_channels,
                kernel_size=2,
                stride=2,
            ),
        )

        # decoder for part 2
        self.dc2 = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=num_feature * 8,
                out_channels=num_feature * 4,
                kernel_size=2,
                stride=2,
            ),
            nn.ConvTranspose2d(
                in_channels=num_feature * 4,
                out_channels=out_channels,
                kernel_size=2,
                stride=2,
            ),
        )

        # decoder for part 1
        self.dc1 = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=num_feature,
                out_channels=out_channels,
                kernel_size=2,
                stride=2,
            ),
        )

        # last layers
        self.final = nn.Sequential(
            nn.Conv2d(
                in_channels=out_channels * 6,
                out_channels=out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
            nn.Conv2d(
                in_channels=out_channels,
                out_channels=out_channels,
                kernel_size=1,
                stride=1,
            ),
        )

    def forward(self, x):
        fea1 = self.block1(self.p1(x))
        fea2 = self.block2(self.p2(fea1))
        fea3 = self.block3(self.p3(fea2))
        fea4 = self.block4(self.p4(fea3))
        fea5 = self.block5(self.p5(fea4))
        fea6 = self.block6(self.p6(fea5))

        pred6 = self.dc6(fea6)
        pred5 = self.dc5(fea5)
        pred4 = self.dc4(fea4)
        pred3 = self.dc3(fea3)
        pred2 = self.dc2(fea2)
        pred1 = self.dc1(fea1)

        pred_join = torch.cat([pred1, pred2, pred3, pred4, pred5, pred6], dim=1)
        pred_final = self.final(pred_join)

        return pred_final


"""
class ConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride, padding=0):
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )
        self.bn = nn.BatchNorm2d(out_channels)

    def forward(self, x):
        return F.relu(self.bn(self.conv(x)), inplace=True)


class InputBlock(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2, padding=0)
        self.conv_1 = ConvBlock(
            in_channels, out_channels // 4, kernel_size=3, stride=1, padding=1
        )
        self.conv_2 = ConvBlock(
            out_channels // 4, out_channels, kernel_size=3, stride=1, padding=1
        )

    def forward(self, x):
        x = self.pool(x)
        x = self.conv_1(x)
        x = self.conv_2(x)
        return x


class Bottleneck(nn.Module):
    def __init__(self, in_channels, bottleneck_channels, out_channels):
        super().__init__()
        self.conv1 = nn.Conv2d(
            in_channels,
            bottleneck_channels,
            kernel_size=1,
            padding=0,
            stride=1,
            bias=True,
        )
        self.bn1 = nn.BatchNorm2d(bottleneck_channels)
        self.conv2 = nn.Conv2d(
            bottleneck_channels,
            bottleneck_channels,
            kernel_size=3,
            padding=1,
            stride=1,
            bias=True,
        )
        self.bn2 = nn.BatchNorm2d(bottleneck_channels)
        self.conv3 = nn.Conv2d(
            bottleneck_channels,
            out_channels,
            kernel_size=1,
            padding=0,
            stride=1,
            bias=True,
        )
        self.bn3 = nn.BatchNorm2d(out_channels)
        self.conv_residual = nn.Conv2d(
            in_channels, out_channels, kernel_size=1, padding=0, stride=1, bias=True
        )

    def forward(self, x):
        z = F.relu(self.bn1(self.conv1(x)), inplace=True)
        z = F.relu(self.bn2(self.conv2(z)), inplace=True)
        z = F.relu(self.bn3(self.conv3(z)), inplace=True)
        z = z + self.conv_residual(x)
        return z


class UpsamplingBlock(nn.Module):
    def __init__(self, out_channels_for_encoder, n_times_to_upsample):
        super().__init__()
        up_convs = []
        initial_in_channel = out_channels_for_encoder[n_times_to_upsample - 1]
        out_channels_for_conv_trans = [
            ch // 4 for ch in out_channels_for_encoder[:n_times_to_upsample][::-1]
        ]

        for i in range(n_times_to_upsample):
            if i == 0:
                up_convs.append(
                    nn.ConvTranspose2d(
                        initial_in_channel,
                        out_channels_for_conv_trans[i],
                        kernel_size=2,
                        stride=2,
                    )
                )
            else:
                up_convs.append(
                    nn.ConvTranspose2d(
                        out_channels_for_conv_trans[i - 1],
                        out_channels_for_conv_trans[i],
                        kernel_size=2,
                        stride=2,
                    )
                )
        self.up_convs = nn.ModuleList(up_convs)

    def forward(self, x):
        for up_conv in self.up_convs:
            x = up_conv(x)
        return x


class IntermediateFeatureMap(nn.Module):
    def __init__(self, in_channels, bottleneck_channels, out_channels):
        super().__init__()
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2, padding=0)
        self.bottleneck_1 = Bottleneck(in_channels, bottleneck_channels, in_channels)
        self.bottleneck_2 = Bottleneck(in_channels, bottleneck_channels, out_channels)

    def forward(self, x):
        x = self.pool(x)
        x = self.bottleneck_1(x)
        x = self.bottleneck_2(x)
        return x


class SuggestiveAnnotationModel(nn.Module):
    def __init__(
        self, in_channels, num_classes, out_channels=[256, 256, 512, 1024, 1024, 1024]
    ):
        super().__init__()
        encoder_blocks = []
        decoder_blocks = []
        for i, output_channel in enumerate(out_channels):
            if i == 0:
                encoder_blocks.append(
                    InputBlock(in_channels=in_channels, out_channels=output_channel)
                )
            else:
                encoder_blocks.append(
                    IntermediateFeatureMap(
                        in_channels=out_channels[i - 1],
                        bottleneck_channels=output_channel // 4,
                        out_channels=output_channel,
                    )
                )
            decoder_blocks.append(
                UpsamplingBlock(
                    out_channels_for_encoder=out_channels, n_times_to_upsample=i + 1
                )
            )
        self.encoder_blocks = nn.ModuleList(encoder_blocks)
        self.decoder_blocks = nn.ModuleList(decoder_blocks)
        self.out_conv_1 = ConvBlock(
            in_channels=(out_channels[0] // 4) * len(out_channels),
            out_channels=num_classes,
            kernel_size=3,
            stride=1,
            padding=1,
        )
        self.out_conv_2 = ConvBlock(
            in_channels=num_classes, out_channels=num_classes, kernel_size=1, stride=1
        )

    def forward(self, x, with_image_embedding=False):
        encoder_outputs = []
        for encoder_block in self.encoder_blocks:
            x = encoder_block(x)
            encoder_outputs.append(x)

        decoder_outputs = []
        for encoder_output, decoder_block in zip(encoder_outputs, self.decoder_blocks):
            decoder_output = decoder_block(encoder_output)
            decoder_outputs.append(decoder_output)

        concat = torch.cat(decoder_outputs, 1)
        out = self.out_conv_1(concat)
        out = self.out_conv_2(out)

        if with_image_embedding:
            *_, last_encoder_output = encoder_outputs
            last_encoder_output = last_encoder_output.view(
                *last_encoder_output.shape[:2], -1
            )
            image_embeddings = last_encoder_output.mean(-1)
            return out, image_embeddings
        else:
            return out
"""



import numpy as np
from typing import Dict
from pathlib import Path
from random import randint
import lightning as pl
import torch
from bioio.writers import OmeTiffWriter

from mmv_im2im.utils.misc import (
    parse_config,
    parse_config_func,
    parse_config_func_without_params,
)
from mmv_im2im.utils.model_utils import init_weights


class Model(pl.LightningModule):
    def __init__(self, model_info_xx: Dict, train: bool = True, verbose: bool = False):
        super().__init__()
        self.net = parse_config(model_info_xx.net)
        # if "sliding_window_params" in model_info_xx:
        #    self.sliding_window = model_info_xx["sliding_window_params"]
        # else:
        #    self.sliding_window = None

        init_weights(self.net, init_type="kaiming")

        self.model_info = model_info_xx
        self.verbose = verbose
        self.weighted_loss = False
        self.seg_flag = False
        if train:
            if "use_costmap" in model_info_xx.criterion[
                "params"
            ] and model_info_xx.criterion["params"].pop("use_costmap"):
                self.weighted_loss = True
            self.criterion = parse_config(model_info_xx.criterion)
            self.optimizer_func = parse_config_func(model_info_xx.optimizer)

            if (
                model_info_xx.model_extra is not None
                and "debug_segmentation" in model_info_xx.model_extra
                and model_info_xx.model_extra["debug_segmentation"]
            ):
                self.seg_flag = True

    def configure_optimizers(self):
        optimizer = self.optimizer_func(self.parameters())
        if self.model_info.scheduler is None:
            return optimizer
        else:
            scheduler_func = parse_config_func_without_params(self.model_info.scheduler)
            lr_scheduler = scheduler_func(
                optimizer, **self.model_info.scheduler["params"]
            )
            return {
                "optimizer": optimizer,
                "lr_scheduler": {
                    "scheduler": lr_scheduler,
                    "monitor": "val_loss",
                    "interval": "epoch",
                    "frequency": 1,
                    "strict": True,
                },
            }

    def prepare_batch(self, batch):
        return

    def forward(self, x):
        return self.net(x)

    def run_step(self, batch, validation_stage):
        x = batch["IM"]
        y = batch["GT"]
        if "CM" in batch.keys():
            assert (
                self.weighted_loss
            ), "Costmap is detected, but no use_costmap param in criterion"
            cm = batch["CM"]

        # only for badly formated data file
        if x.size()[-1] == 1:
            x = torch.squeeze(x, dim=-1)
            y = torch.squeeze(y, dim=-1)

        y_hat = self(x)

        if isinstance(self.criterion, torch.nn.CrossEntropyLoss):
            # in case of CrossEntropy related error
            # see: https://discuss.pytorch.org/t/runtimeerror-expected-object-of-scalar-type-long-but-got-scalar-type-float-when-using-crossentropyloss/30542  # noqa E501
            y = torch.squeeze(y, dim=1)  # remove C dimension

        if self.weighted_loss:
            loss = self.criterion(y_hat, y, cm)
        else:
            loss = self.criterion(y_hat, y)

        return loss, y_hat

    def training_step(self, batch, batch_idx):
        loss, y_hat = self.run_step(batch, validation_stage=False)
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        if self.verbose and batch_idx == 0:
            src = batch["IM"]
            tar = batch["GT"]

            # check if the log path exists, if not create one
            save_path = Path(self.trainer.log_dir)
            save_path.mkdir(parents=True, exist_ok=True)

            # check if need to use softmax
            if self.seg_flag:
                act_layer = torch.nn.Softmax(dim=1)
                yhat_act = act_layer(y_hat)
            else:
                yhat_act = y_hat

            src_out = np.squeeze(src[0,].detach().cpu().numpy()).astype(float)
            tar_out = np.squeeze(tar[0,].detach().cpu().numpy()).astype(float)
            prd_out = np.squeeze(yhat_act[0,].detach().cpu().numpy()).astype(float)

            def get_dim_order(arr):
                dims = len(arr.shape)
                return {2: "YX", 3: "ZYX", 4: "CZYX"}.get(dims, "YX")

            rand_tag = randint(1, 1000)
            out_fn = save_path / f"epoch_{self.current_epoch}_src_{rand_tag}.tiff"
            OmeTiffWriter.save(src_out, out_fn, dim_order=get_dim_order(src_out))
            out_fn = save_path / f"epoch_{self.current_epoch}_tar_{rand_tag}.tiff"
            OmeTiffWriter.save(tar_out, out_fn, dim_order=get_dim_order(tar_out))
            out_fn = save_path / f"epoch_{self.current_epoch}_prd_{rand_tag}_.tiff"
            OmeTiffWriter.save(prd_out, out_fn, dim_order=get_dim_order(prd_out))

        return loss

    def validation_step(self, batch, batch_idx):
        loss, y_hat = self.run_step(batch, validation_stage=True)
        self.log(
            "val_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        return loss
